#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

""" Interface with neuprint. This module is wrapping neuprint-python
(https://github.com/connectome-neuprint/neuprint-python) and adding some
navis-specific functions.
"""

import trimesh

from urllib.parse import urlparse
from textwrap import dedent

from concurrent.futures import ThreadPoolExecutor, as_completed

try:
    from neuprint import *
    # remove neuprint's own fetch_skeleton function to avoid confusion
    del fetch_skeleton  # noqa
    from neuprint.client import inject_client
except ModuleNotFoundError:
    msg = dedent("""
          neuprint library not found. Please install using pip:

                pip install neuprint-python

          """)
    raise ModuleNotFoundError(msg)
except BaseException:
    raise

import io

from requests.exceptions import HTTPError

import numpy as np
import pandas as pd

from .. import config, utils

from ..core import Volume, TreeNeuron, MeshNeuron, NeuronList
from ..graph import neuron2KDTree
from ..morpho import subset_neuron

logger = config.get_logger(__name__)

# Define some integer types
int_types = (int, np.integer)


@inject_client
def fetch_roi(roi, *, client=None):
    """Fetch given ROI.

    Parameters
    ----------
    roi :           str
                    Name of an ROI.
    client :        neuprint.Client, optional
                    If `None` will try using global client.

    Returns
    -------
    navis.Volume

    """
    if not isinstance(roi, str):
        raise TypeError(f'Expect ROI name as string, got "{type(roi)}"')

    # Fetch data
    data = client.fetch_roi_mesh(roi, export_path=None)

    # Turn into file-like object
    f = io.StringIO(data.decode())

    # Parse with trimesh
    ob = trimesh.load_mesh(f, file_type='obj')

    return Volume.from_object(ob, name=roi)


@inject_client
def fetch_mesh_neuron(x, *, lod=1, with_synapses=False, missing_mesh='raise',
                      parallel=True, max_threads=5, seg_source=None,
                      client=None, **kwargs):
    """Fetch neuron meshes as navis.MeshNeuron.

    Requires additional packages depending on the mesh source.

    For DVID you need [`dvid-tools`](https://github.com/flyconnectome/dvid_tools):

        ``` shell
        pip3 install dvidtools
        ```

    For everything else you need [cloudvolume](https://github.com/seung-lab/cloud-volume):

        ``` shell
        pip3 install cloud-volume
        ```


    Parameters
    ----------
    x :             str | int | list-like | pandas.DataFrame | SegmentCriteria
                    Body ID(s). Multiple IDs can be provided as list-like or
                    DataFrame with "bodyId" or "bodyid" column.
    lod :           int
                    Level of detail. Higher `lod` = coarser. Ignored if mesh
                    source does not support LODs (e.g. for DVID).
    with_synapses : bool, optional
                    If True will download and attach synapses as `.connectors`.
    missing_mesh :  'raise' | 'warn' | 'skip'
                    What to do if no mesh is found for a given body ID:

                        "raise" (default) will raise an exception
                        "warn" will throw a warning but continue
                        "skip" will skip without any message

    parallel :      bool
                    If True, will use parallel threads to fetch data.
    max_threads :   int
                    Max number of parallel threads to use.
    seg_source :    str | cloudvolume.CloudVolume, optional
                    Use this to override the segmentation source specified by
                    neuPrint.
    client :        neuprint.Client, optional
                    If `None` will try using global client.
    **kwargs
                    Will be passed to `cloudvolume.CloudVolume`.

    Returns
    -------
    navis.Neuronlist
                    Containing [`navis.MeshNeuron`][]. Note that meshes are
                    resized to raw voxel size to match other spatial data from
                    neuprint (synapses, skeletons, etc).

    """
    if isinstance(x, pd.DataFrame):
        if 'bodyId' in x.columns:
            x = x['bodyId'].values
        elif 'bodyid' in x.columns:
            x = x['bodyid'].values
        else:
            raise ValueError('DataFrame must have "bodyId" column.')

    # Extract source
    if not seg_source:
        seg_source = get_seg_source(client=client)

    if not seg_source:
        raise ValueError('Segmentation source could not be automatically '
                         'determined. Please provide via `seg_source`.')

    if isinstance(seg_source, str) and seg_source.startswith('dvid'):
        try:
            import dvid as dv
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                'This looks like a DVID mesh source. For this we '
                'need the `dvid-tools` library:\n'
                '  pip3 install dvidtools -U')
        o = urlparse(seg_source.replace('dvid://', ''))
        server = f'{o.scheme}://{o.netloc}'
        node = o.path.split('/')[1]

        if lod is not None:
                logger.warning(
                    'This dataset does not support LODs. '
                    'Will ignore the `lod` argument. '
                    'You can silence this warning by setting `lod=None`.')
                lod = None
    else:
        try:
            from cloudvolume import CloudVolume
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "You need to install the `cloudvolume` library"
                "to fetch meshes from this mesh source:\n"
                "  pip3 install cloud-volume -U")
        # Initialize volume
        if isinstance(seg_source, CloudVolume):
            vol = seg_source
        else:
            defaults = dict(use_https=True, progress=False)
            defaults.update(kwargs)
            vol = CloudVolume(seg_source, **defaults)

            # Check if vol.mesh.get has a lod argument
            if lod is not None and 'lod' not in vol.mesh.get.__code__.co_varnames:
                logger.warning(
                    'This dataset does not have multi-resolution meshes and '
                    'the `lod` parameter will be ignored. '
                    'You can silence this warning by setting `lod=None`.')
                lod = None

    if isinstance(x, NeuronCriteria):
        query = x
        wanted_ids = None
    else:
        query = NeuronCriteria(bodyId=x, client=client)
        wanted_ids = utils.make_iterable(x)

    # Fetch names, etc
    meta, roi_info = fetch_neurons(query, client=client)

    if meta.empty:
        raise ValueError('No neurons matching the given criteria found!')
    elif not isinstance(wanted_ids, type(None)):
        miss = wanted_ids[~np.isin(wanted_ids, meta.bodyId.values)]
        if len(miss):
            logger.warning(f'Skipping {len(miss)} body IDs that were not found: '
                           f'{", ".join(miss.astype(str))}')

    # Make sure there is a somaLocation and somaRadius column
    if 'somaLocation' not in meta.columns:
        meta['somaLocation'] = None
    if 'somaRadius' not in meta.columns:
        meta['somaRadius'] = None

    if isinstance(seg_source, str) and seg_source.startswith('dvid'):
        # Fetch the meshes
        nl = dv.get_meshes(meta.bodyId.values,
                           on_error=missing_mesh,
                           output='navis',
                           progress=meta.shape[0] > 1 and not config.pbar_hide,
                           max_threads=1 if not parallel else max_threads,
                           server=server,
                           node=node)
    else:
        nl = []
        with ThreadPoolExecutor(max_workers=1 if not parallel else max_threads) as executor:
            futures = {}
            for r in meta.itertuples():
                f = executor.submit(__fetch_mesh,
                                    r.bodyId,
                                    vol=vol,
                                    lod=lod,
                                    missing_mesh=missing_mesh)
                futures[f] = r.bodyId

            with config.tqdm(desc='Fetching',
                             total=len(futures),
                             leave=config.pbar_leave,
                             disable=meta.shape[0] == 1 or config.pbar_hide) as pbar:
                for f in as_completed(futures):
                    bodyId = futures[f]
                    pbar.update(1)
                    try:
                        nl.append(f.result())
                    except Exception as exc:
                        print(f'{bodyId} generated an exception:', exc)

    nl = NeuronList(nl)

    # Add meta data
    instances = meta.set_index('bodyId').instance.to_dict()
    sizes = meta.set_index('bodyId')['size'].to_dict()
    status = meta.set_index('bodyId').status.to_dict()
    statuslabel = meta.set_index('bodyId').statusLabel.to_dict()
    somalocs = meta.set_index('bodyId').somaLocation.to_dict()
    radii = meta.set_index('bodyId').somaRadius.to_dict()

    for n in nl:
        n.name = instances[n.id]
        n.status = status[n.id]
        n.statusLabel = statuslabel[n.id]
        n.n_voxels = sizes[n.id]
        n.somaLocation = somalocs[n.id]

        # Meshes come out in units (e.g. nanometers) but most other data (synapses,
        # skeletons, etc) come out in voxels, we will therefore scale meshes to voxels
        n.vertices /= np.array(client.meta['voxelSize']).reshape(1, 3)
        n.units=f'{client.meta["voxelSize"][0]} {client.meta["voxelUnits"]}'

        if n.somaLocation:
            if radii[n.id]:
                n.soma_radius = radii[n.id] / n.units.to('nm').magnitude
            else:
                n.soma_radius = None
            n.soma_pos = n.somaLocation

    if with_synapses:
        # Fetch synapses
        syn = fetch_synapses(meta.bodyId.values,
                             synapse_criteria=SynapseCriteria(primary_only=True, client=client),
                             client=client)

        for n in nl:
            this_syn = syn[syn.bodyId == n.id]
            if not this_syn.empty:
                # Keep only relevant columns
                n.connectors = syn[['type', 'x', 'y', 'z', 'roi', 'confidence']]

    # Make an effort to retain the original order
    if not isinstance(x, NeuronCriteria) and not nl.empty:
        nl = nl.idx[np.asarray(x)[np.isin(x, nl.id)]]

    return nl


def __fetch_mesh(bodyId, *, vol, lod, missing_mesh='raise'):
    """Fetch a single mesh (+ synapses) and construct navis MeshNeuron."""
    # Fetch mesh
    import cloudvolume
    try:
        if lod is None:
            mesh = vol.mesh.get(bodyId)
        else:
            mesh = vol.mesh.get(bodyId, lod=lod)
    except cloudvolume.exceptions.MeshDecodeError as err:
        if 'not found' in str(err):
            if missing_mesh in ['warn', 'skip']:
                if missing_mesh == 'warn':
                    logger.warning(f'No mesh found for {r.bodyId}')
                return
            else:
                raise
        else:
            raise

    # Make sure we don't pass through a {bodyId: MeshObject} dictionary
    if isinstance(mesh, dict):
        mesh = mesh[bodyId]

    n = MeshNeuron(mesh)
    n.lod = lod
    n.id = bodyId

    return n


@inject_client
def fetch_skeletons(x, *, with_synapses=False, heal=False, missing_swc='raise',
                    parallel=True, max_threads=5, client=None):
    """Fetch neuron skeletons as navis.TreeNeurons.

    Notes
    -----
    Synapses will be attached to the closest node in the skeleton.

    Parameters
    ----------
    x :             str | int | list-like | pandas.DataFrame | SegmentCriteria
                    Body ID(s). Multiple Ids can be provided as list-like or
                    DataFrame with "bodyId"  or "bodyid" column.
    with_synapses : bool, optional
                    If True will also attach synapses as `.connectors`.
    heal :          bool | int | float, optional
                    If True, will automatically heal fragmented skeletons using
                    neuprint-python's `heal_skeleton` function. Pass a float
                    or an int to limit the max distance at which nodes are
                    allowed to be re-connected (requires neuprint-python >= 0.4.11).
    missing_swc :   'raise' | 'warn' | 'skip'
                    What to do if no skeleton is found for a given body ID:

                      - "raise" (default) will raise an exception
                      - "warn" will throw a warning but continue
                      - "skip" will skip without any message

    parallel :      bool
                    If True, will use parallel threads to fetch data.
    max_threads :   int
                    Max number of parallel threads to use.
    client :        neuprint.Client, optional
                    If `None` will try using global client.

    Returns
    -------
    navis.Neuronlist

    """
    if isinstance(x, pd.DataFrame):
        if 'bodyId' in x.columns:
            x = x['bodyId'].values
        elif 'bodyid' in x.columns:
            x = x['bodyid'].values
        else:
            raise ValueError('DataFrame must have "bodyId" column.')

    if isinstance(x, NeuronCriteria):
        query = x
        wanted_ids = None
    else:
        query = NeuronCriteria(bodyId=x, client=client)
        wanted_ids = utils.make_iterable(x)

    # Fetch names, etc
    meta, roi_info = fetch_neurons(query, client=client)

    if meta.empty:
        raise ValueError('No neurons matching the given criteria found!')
    elif not isinstance(wanted_ids, type(None)):
        miss = wanted_ids[~np.isin(wanted_ids, meta.bodyId.values)]
        if len(miss):
            logger.warning(f'Skipping {len(miss)} body IDs that were not found: '
                           f'{", ".join(miss.astype(str))}')

    # Make sure there is a somaLocation and somaRadius column
    if 'somaLocation' not in meta.columns:
        meta['somaLocation'] = None
    if 'somaRadius' not in meta.columns:
        meta['somaRadius'] = None

    nl = []
    with ThreadPoolExecutor(max_workers=1 if not parallel else max_threads) as executor:
        futures = {}
        for r in meta.itertuples():
            f = executor.submit(__fetch_skeleton,
                                r,
                                client=client,
                                with_synapses=with_synapses,
                                missing_swc=missing_swc,
                                heal=heal)
            futures[f] = r.bodyId

        with config.tqdm(desc='Fetching',
                         total=meta.shape[0],
                         leave=config.pbar_leave,
                         disable=meta.shape[0] == 1 or config.pbar_hide) as pbar:
            for f in as_completed(futures):
                bodyId = futures[f]
                pbar.update(1)
                try:
                    nl.append(f.result())
                except Exception as exc:
                    print(f'{bodyId} generated an exception:', exc)

    nl = NeuronList(nl)

    # Make an effort to retain the original order
    if not isinstance(x, NeuronCriteria) and not nl.empty:
        nl = nl.idx[np.asarray(x)[np.isin(x, nl.id)]]

    return nl


def __fetch_skeleton(r, client, with_synapses=True, missing_swc='raise',
                     heal=False, max_distance=None):
    """Fetch a single skeleton + synapses and construct navis TreeNeuron."""
    # Fetch skeleton SWC
    try:
        data = client.fetch_skeleton(r.bodyId, format='pandas', heal=heal)
    except HTTPError as err:
        if err.response.status_code == 400:
            if missing_swc in ['warn', 'skip']:
                if missing_swc == 'warn':
                    logger.warning(f'No SWC found for {r.bodyId}')
                return
            else:
                raise
        else:
            raise

    # Generate neuron
    # Note that we are currently assuming that the x/y/z data is isotropic
    n = TreeNeuron(data,
                   units=f'{client.meta["voxelSize"][0]} {client.meta["voxelUnits"]}')

    # Reduce precision
    n._nodes = n._nodes.astype({'node_id': np.int32, 'parent_id': np.int32,
                                'x': np.float32, 'y': np.float32,
                                'z': np.float32, 'radius': np.float32})

    # Add some missing meta data
    n.id = r.bodyId

    if hasattr(r, 'instance'):
        n.name = r.instance

    n.n_voxels = r.size
    n.status = r.status

    # Make KDE tree for NN
    if r.somaLocation or with_synapses:
        tree = neuron2KDTree(n, data='nodes')

    # Set soma
    if r.somaLocation:
        d, i = tree.query([r.somaLocation])
        n.soma = int(n.nodes.iloc[i[0]].node_id)
        n.soma_radius = r.somaRadius if r.somaRadius else 'radius'
    else:
        n.soma = None

    if with_synapses:
        # Fetch synapses
        syn = fetch_synapses(r.bodyId,
                             synapse_criteria=SynapseCriteria(primary_only=True, client=client),
                             client=client)

        if not syn.empty:
            # Process synapses
            syn['connector_id'] = syn.index.values
            locs = syn[['x', 'y', 'z']].values
            d, i = tree.query(locs)

            syn['node_id'] = n.nodes.iloc[i].node_id.values
            syn['x'] = locs[:, 0]
            syn['y'] = locs[:, 1]
            syn['z'] = locs[:, 2]

            # Keep only relevant columns
            syn = syn[['connector_id', 'node_id', 'type',
                       'x', 'y', 'z', 'roi', 'confidence']]

            # Manually make the "roi" column of the synapse table into a
            # categorical to save some memory
            syn['roi'] = syn.roi.astype('category')

            n.connectors = syn

    return n


def remove_soma_hairball(x: 'core.TreeNeuron',
                         radius: float = 500,
                         inplace: bool = False):
    """Remove hairball around soma.

    Parameters
    ----------
    x :         core.TreeNeuron
    radius :    float
                Radius around the soma to check for hairball

    Returns
    -------
    TreeNeuron
                If inplace=False.
    """
    if not inplace:
        x = x.copy()
    if not x.soma:
        if not inplace:
            return x
        return
    # Get all nodes within given radius of soma nodes
    soma_loc = x.nodes.set_index('node_id').loc[[x.soma],
                                                ['x', 'y', 'z']].values
    tree = neuron2KDTree(x)
    dist, ix = tree.query(soma_loc, k=x.n_nodes, distance_upper_bound=radius)

    # Subset to nodes within range
    to_check = set(list(ix[0, dist[0, :] <= radius]))

    # Get the segments that have nodes in the soma
    segs = [seg for seg in x.segments if set(seg) & to_check]

    # Unless these segments end in a root node, we will keep the last node
    # (which will be a branch point)
    segs = [s[:-1] if s[-1] not in x.root else s for s in segs]

    # This is already sorted by length -> we will keep the first (i.e. longest)
    # segment and remove the rest
    to_remove = [n for s in segs[1:] for n in s]

    to_keep = x.nodes.loc[~x.nodes.node_id.isin(to_remove), 'node_id'].values

    # Move soma if required
    if x.soma in to_remove:
        x.soma = list(to_check & set(to_keep))[0]

    subset_neuron(x, to_keep, inplace=True)

    if not inplace:
        return x


@inject_client
def get_seg_source(*, client=None):
    """Get segmentation source for given client+dataset."""
    # First try to fetch the scene for the neuroglancer
    url = f'{client.server}/api/npexplorer/nglayers/{client.dataset}.json'

    r = client.session.get(url)
    try:
        r.raise_for_status()
        scene = r.json()
        segs = [s for s in scene['layers'] if s.get('type') == 'segmentation']
    except BaseException:
        segs = []

    # If we didn't find a `dataset.json`, will check the client's meta data for a seg source
    if not segs:
        segs = [s for s in client.meta['neuroglancerMeta'] if s.get('dataType') == 'segmentation']

    if not len(segs):
        return None

    # Check if any segmentation source matches our dataset exactly
    named_segs = [s for s in segs if s.get('name') == client.dataset]
    if len(named_segs):
        segs = named_segs

    # If there are multiple segmentation layers, select the first entry
    seg_source = segs[0]['source']

    # If there are multiple segmentation sources for
    # the layer we picked, select the first source.
    if isinstance(seg_source, list):
        seg_source = seg_source[0]

    # If it's a dict like {'source': url, 'subsources'...},
    # select the url.
    if isinstance(seg_source, dict):
        seg_source = seg_source['url']

    if not isinstance(seg_source, str):
        e = f"Could not understand segmentation source: {seg_source}"
        raise RuntimeError(e)

    if len(segs) > 1:
        logger.warning(f'{len(segs)} segmentation sources found. Using the '
                       f'first entry: "{seg_source}"')

    return seg_source



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

"""This module contains functions for intersections."""

import numpy as np

from typing import Sequence, Optional

from ..core import Volume

try:
    from pyoctree import pyoctree
except ModuleNotFoundError:
    pyoctree = None

try:
    import ncollpyde
except ModuleNotFoundError:
    ncollpyde = None


def in_volume_ncoll(points: np.ndarray,
                    volume: Volume,
                    n_rays: Optional[int] = 3) -> Sequence[bool]:
    """Use ncollpyde to test if points are within a given volume."""
    if isinstance(n_rays, type(None)):
        n_rays = 3

    if not isinstance(n_rays, (int, np.integer)):
        raise TypeError(f'n_rays must be integer, got "{type(n_rays)}"')

    if n_rays <= 0:
        raise ValueError('n_rays must be > 0')

    coll = ncollpyde.Volume(volume.vertices, volume.faces, n_rays=n_rays)

    return coll.contains(points)


def in_volume_pyoc(points: np.ndarray,
                   volume: Volume,
                   n_rays: Optional[int] = 1) -> Sequence[bool]:
    """Use pyoctree's raycasting to test if points are within a given volume."""
    if isinstance(n_rays, type(None)):
        n_rays = 1

    if not isinstance(n_rays, (int, np.integer)):
        raise TypeError(f'n_rays must be integer, got "{type(n_rays)}"')

    if n_rays <= 0:
        raise ValueError('n_rays must be > 0')

    tree = getattr(volume, 'pyoctree', None)

    if not tree:
        # Create octree from scratch
        tree = pyoctree.PyOctree(np.array(volume.vertices, dtype=float, order='C'),
                                 np.array(volume.faces, dtype=np.int32, order='C')
                                 )
        volume.pyoctree = tree  # type: ignore

    # Get min max of volume
    mx = np.array(volume.vertices).max(axis=0)
    mn = np.array(volume.vertices).min(axis=0)
    dm = mx - mn

    # Remove points outside of bounding box
    is_out = (points > mx).any(axis=1) | (points < mn).any(axis=1)

    # Cast rays
    # There is no point of vectorizing this because pyoctree's rayIntersection
    # does only take a single ray at a time...
    for i in range(n_rays):
        # Process only point that we think could be in
        in_points = points[~is_out]

        # If no in points left, break out
        if in_points.size == 0:
            break

        # Pick a random point inside the volumes bounding box as origin
        origin = np.random.rand(3) * dm + mn

        # Generate ray point list:
        rayPointList = np.array([[origin, p] for p in in_points],
                                dtype=np.float32)

        # Get intersections and extract coordinates of intersection
        intersections = [np.array([i.p for i in tree.rayIntersection(ray)]) for ray in rayPointList]

        # In a few odd cases we can get the multiple intersections at the exact
        # same coordinate (something funny with the faces).
        unique_int = [np.unique(np.round(i), axis=0) if np.any(i) else i for i in intersections]

        # Unfortunately rays are bidirectional -> we have to filter intersections
        # to those that occur "above" the point we are querying
        unilat_int = [i[i[:, 2] >= p] if np.any(i) else i for i, p in zip(unique_int, in_points[:, 2])]

        # Count intersections
        int_count = [i.shape[0] for i in unilat_int]

        # Get even (= outside volume) numbers of intersections
        is_even = np.remainder(int_count, 2) == 0

        # Set outside points
        is_out[~is_out] = is_even

    return ~is_out



"""
Code to execute code tutorial notebooks in /docs/examples/.

This will not be run through pytest but is executed in a separate CI job.

A couple notes:
 - the tutorials require a number of extra dependencies and data files to be present
   check out the test-tutorials.yml workflow to see how this is set up.
 - the MICrONS tutorial occasionally fails because the CAVE backend throws an error
   (e.g. during the materialization)
 - Github runners appear to have 4 CPUs - so should be good to go
"""

import os
import sys
import navis
import warnings

import matplotlib.pyplot as plt

from pathlib import Path
from contextlib import contextmanager


@contextmanager
def suppress_stdout():
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        try:
            yield
        finally:
            sys.stdout = old_stdout

# Silence logging
navis.config.logger.setLevel("ERROR")

# Hide pbars
navis.set_pbars(hide=True)

# Silence warnings
warnings.filterwarnings("ignore")


if __name__ == "__main__":
    # Recursively search for notebooks
    path = Path(__file__).parent.parent / "docs/examples"

    files = list(path.rglob("*.py"))
    for i, file in enumerate(files):
        if not file.is_file():
            continue
        if file.name.startswith('zzz'):
            continue

        # Note: we're using `exec` here instead of e.g. `subprcoess.run` because we need to avoid
        # the "An attempt has been made to start a new process before the current process has
        # finished its bootstrapping phase" error that occurs when using multiprocessing with "spawn"
        # from a process where it's not wrapped in an `if __name__ == "__main__":` block.
        print(f"Executing {file.name} [{i+1}/{len(files)}]... ", end="", flush=True)
        with suppress_stdout():
            os.chdir(file.parent)
            exec(open(file.name).read())
        print("Done.", flush=True)

        # Make sure to close any open figures
        plt.close("all")

    print("All done.")



import pint

import numpy as np
import matplotlib as mpl

from dataclasses import dataclass, field
from typing import Union, List, Tuple, Optional
from typing_extensions import Literal

# Global flag whether to validate settings
VALIDATE_SETTINGS = True


@dataclass
class Settings:
    """Class that works a bit like a dictionary but can validate keys and has some extra features."""

    # We can define synonyms for arguments, so that they can be used interchangeably
    _synonyms: List[Tuple] = field(default_factory=list)
    _name = "Settings"

    def __setattr__(self, key, value, check_valid=False):
        if check_valid and key not in self.properties:
            raise AttributeError(
                f"The '{key}' argument is invalid for {self._name}. Valid arguments are: {', '.join(self.properties)}"
            )
        self.__dict__[key] = value

    def __contains__(self, key):
        return key in self.properties

    @property
    def properties(self):
        return tuple(
            [
                p
                for p in dir(self)
                if not p.startswith("_")
                and (p != "properties")   # we need this to avoid infinite recursion
                and not callable(getattr(self, p, None))
            ]
        )

    def update_settings(self, **kwargs):
        # Deal with synonyms
        for syn in self._synonyms:
            present = [s for s in syn if s in kwargs]
            if len(present) > 1:
                raise ValueError(f"Multiple synonyms for the same argument: {present}")

            for s in syn[1:]:
                if s in kwargs:
                    kwargs[syn[0]] = kwargs.pop(s)

        for k, v in kwargs.items():
            self.__setattr__(k, v, check_valid=VALIDATE_SETTINGS)

        # For method chaining
        return self

    def to_dict(self):
        return {k: v for k, v in self.__dict__.items() if not k.startswith("_")}

    def get(self, key, default=None):
        value = self.__dict__.get(key, default)
        if value is None:
            value = default
        return value

    def pop(self, key, default=None):
        return self.__dict__.pop(key, default)


@dataclass
class BasePlottingSettings(Settings):
    """Plotting parameters common to all functions/backends."""

    _name = "BasePlottingSettings"

    # For TreeNeurons
    soma: bool = True
    radius: bool = False  # True | False | "auto"
    linewidth: float = 1
    linestyle: str = "-"

    # For Dotprops
    dps_scale_vec: float = "auto"

    # All neurons
    connectors: bool = False
    connectors_only: bool = False
    cn_size: Optional[float] = None
    cn_alpha: Optional[float] = None
    cn_layout: dict = field(default_factory=dict)
    cn_colors: dict = field(default_factory=dict)
    cn_mesh_colors: bool = False
    color: Optional[
        Union[
            str,
            Tuple[float, float, float],
            List[Union[str, Tuple[float, float, float]]],
            dict,
        ]
    ] = None
    color_by: Optional[Union[str, np.ndarray, List[np.ndarray]]] = None
    shade_by: Optional[Union[str, np.ndarray, List[np.ndarray]]] = None
    palette: Optional[Union[str, np.ndarray]] = None
    alpha: Optional[float] = None
    vmin: Optional[float] = None
    vmax: Optional[float] = None
    smin: Optional[float] = None
    smax: Optional[float] = None
    norm_global: bool = True

    # Other
    scatter_kws: dict = field(default_factory=dict)

    _synonyms: List[Tuple] = field(
        default_factory=lambda: [
            ("linestyle", "ls"),
            ("linewidth", "lw"),
            ("color", "colors", "c"),
        ]
    )


@dataclass
class Matplotlib2dSettings(BasePlottingSettings):
    """Additional plotting parameters for Matplotlib 2d backend."""

    _name = "matplotlib backend"

    method: Literal["2d", "3d", "3d_complex"] = "2d"
    group_neurons: bool = False
    autoscale: bool = True
    orthogonal: bool = True
    scalebar: Union[int, float, str, pint.Quantity] = False
    volume_outlines: bool = False
    rasterize: bool = False
    view: Tuple[str, str] = ("x", "y")
    figsize: Optional[Tuple[float, float]] = None
    ax: Optional[mpl.axes.Axes] = None
    mesh_shade: bool = False
    non_view_axes3d: Literal["hide", "show", "fade"] = "hide"

    depth_coloring: bool = False
    depth_scale: bool = True


@dataclass
class PlotlySettings(BasePlottingSettings):
    """Additional plotting parameters for Plotly backend."""

    _name = "plotly backend"

    fig: Optional[Union["plotly.Figure", dict]] = None
    inline: bool = True
    title: Optional[str] = None
    fig_autosize: bool = True
    hover_name: Optional[str] = False
    hover_id: bool = False
    legend: bool = True
    legend_orientation: Literal["h", "v"] = "v"
    legend_group: Optional[str] = None
    volume_legend: bool = False
    width: Optional[int] = None
    height: Optional[int] = 600
    linewidth: Optional[float] = None  # for plotly, linewidth 1 is too thin, we default to 3 in graph_objs.py
    linestyle: str = "-"


@dataclass
class VispySettings(BasePlottingSettings):
    """Additional plotting parameters for Vispy backend."""

    _name = "vispy backend"

    clear: bool = False
    center: bool = True
    combine: bool = False
    title: Optional[str] = None
    viewer: Optional["navis.Viewer"] = None
    shininess: float = 0
    shading: str = "smooth"
    size: Optional[Tuple[int, int]] = (800, 600)
    show: bool = True
    name: Optional[str] = None


@dataclass
class OctarineSettings(BasePlottingSettings):
    """Additional plotting parameters for Octarine backend."""

    _name = "octarine backend"

    clear: bool = False
    center: bool = True
    viewer: Optional[Union["octarine.Viewer", Literal["new"]]] = None
    random_ids: bool = False
    camera: Literal["ortho", "perspective"] = "ortho"
    control: Literal["trackball", "panzoom", "fly", "orbit"] = "trackball"
    show: bool = True
    size: Optional[Tuple[int, int]] = None
    offscreen: bool = False
    spacing: Optional[Tuple[float, float, float]] = None

    # These are viewer-specific settings that we must not pass to the plotting
    # function
    _viewer_settings: tuple[str] = (
        "clear",
        "center",
        "viewer",
        "camera",
        "control",
        "show",
        "size",
        "offscreen",
        "scatter_kws",
        "spacing"
    )


@dataclass
class K3dSettings(BasePlottingSettings):
    """Additional plotting parameters for K3d backend."""

    _name = "k3d backend"

    height: int = 600
    inline: bool = True
    legend_group: Optional[str] = None
    plot: Optional["k3d.Plot"] = None



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import k3d
import uuid
import warnings

import numpy as np
import pandas as pd
import trimesh as tm

from ..colors import vertex_colors, eval_color, color_to_int
from ..plot_utils import segments_to_coords, fibonacci_sphere
from ... import core, utils, config, conversion

logger = config.get_logger(__name__)

__all__ = ["neuron2k3d", "scatter2k3d", "dotprops2k3d", "voxel2k3d", "volume2k3d"]

# Generate sphere for somas
fib_points = fibonacci_sphere(samples=30)


def neuron2k3d(x, colormap, settings):
    """Convert neurons to k3d objects."""
    if isinstance(x, core.BaseNeuron):
        x = core.NeuronList(x)
    elif not isinstance(x, core.NeuronList):
        raise TypeError('Unable to process data of type "{}"'.format(type(x)))

    # palette = kwargs.get("palette", None)
    # color_by = kwargs.get("color_by", None)
    # shade_by = kwargs.get("shade_by", None)
    # lg = kwargs.pop("legend_group", None)

    if settings.color_by is not None:
        if not settings.palette:
            raise ValueError(
                'Must provide `palette` (e.g. "viridis") argument '
                "if using `color_by`"
            )

        colormap = vertex_colors(
            x,
            by=settings.color_by,
            alpha=settings.alpha,
            use_alpha=False,
            palette=settings.palette,
            vmin=settings.vmin,
            vmax=settings.vmax,
            na="raise",
            color_range=255,
        )

    if not isinstance(settings.shade_by, type(None)):
        logger.warning("`shade_by` does not work with the k3d backend")

    cn_lay = config.default_connector_colors.copy()
    cn_lay.update(settings.cn_layout)

    trace_data = []
    _radius_warned = False
    for i, neuron in enumerate(x):
        name = str(getattr(neuron, "name", neuron.id))
        color = colormap[i]

        try:
            # Try converting this neuron's ID
            neuron_id = str(neuron.id)
        except BaseException:
            # If that doesn't work generate a new ID
            neuron_id = str(str(uuid.uuid1()))

        showlegend = True
        label = neuron.label
        if (
            isinstance(settings.legend_group, dict)
            and neuron.id in settings.legend_group
        ):
            # Check if this the first entry for this legendgroup
            label = legendgroup = settings.legend_group[neuron.id]
            for d in trace_data:
                # If it is not the first entry, hide it
                if getattr(d, "legendgroup", None) == legendgroup:
                    showlegend = False
                    break
        elif isinstance(settings.legend_group, str):
            legendgroup = settings.legend_group
        else:
            legendgroup = neuron_id

        if isinstance(neuron, core.TreeNeuron) and settings.radius == "auto":
            # Number of nodes with radii
            n_radii = (neuron.nodes.get("radius", pd.Series([])).fillna(0) > 0).sum()
            # If less than 30% of nodes have a radius, we will fall back to lines
            if n_radii / neuron.nodes.shape[0] < 0.3:
                settings.radius = False

        if isinstance(neuron, core.TreeNeuron) and settings.radius:
            # Warn once if more than 5% of nodes have missing radii
            if not _radius_warned:
                if (
                    (neuron.nodes.radius.fillna(0).values <= 0).sum() / neuron.n_nodes
                ) > 0.05:
                    logger.warning(
                        "Some skeleton nodes have radius <= 0. This may lead to "
                        "rendering artifacts. Set `radius=False` to plot skeletons "
                        "as single-width lines instead."
                    )
                    _radius_warned = True

            _neuron = conversion.tree2meshneuron(
                neuron,
                warn_missing_radii=False,
                radius_scale_factor=settings.get("linewidth", 1),
            )
            _neuron.connectors = neuron.connectors
            neuron = _neuron

            # See if we need to map colors to vertices
            if isinstance(color, np.ndarray) and color.ndim == 2:
                color = color[neuron.vertex_map]

        if not settings.connectors_only:
            if isinstance(neuron, core.TreeNeuron):
                trace_data += skeleton2k3d(
                    neuron,
                    label=label,
                    legendgroup=legendgroup,
                    showlegend=showlegend,
                    color=color,
                    settings=settings,
                )
            elif isinstance(neuron, core.MeshNeuron):
                trace_data += mesh2k3d(
                    neuron,
                    label=label,
                    legendgroup=legendgroup,
                    showlegend=showlegend,
                    color=color,
                    settings=settings,
                )
            elif isinstance(neuron, core.Dotprops):
                trace_data += dotprops2k3d(
                    neuron,
                    label=label,
                    legendgroup=legendgroup,
                    showlegend=showlegend,
                    color=color,
                    settings=settings,
                )
            elif isinstance(neuron, core.VoxelNeuron):
                trace_data += voxel2k3d(
                    neuron,
                    label=label,
                    legendgroup=legendgroup,
                    showlegend=showlegend,
                    color=color,
                    settings=settings,
                )
            else:
                raise TypeError(f'Unable to plot neurons of type "{type(neuron)}"')

        # Add connectors
        if (settings.connectors or settings.connectors_only) and neuron.has_connectors:
            if isinstance(settings.connectors, (list, np.ndarray, tuple)):
                connectors = neuron.connectors[
                    neuron.connectors.type.isin(settings.connectors)
                ]
            elif settings.connectors == "pre":
                connectors = neuron.presynapses
            elif settings.connectors == "post":
                connectors = neuron.postsynapses
            elif isinstance(settings.connectors, str):
                connectors = neuron.connectors[
                    neuron.connectors.type == settings.connectors
                ]
            else:
                connectors = neuron.connectors

            for j, this_cn in connectors.groupby("type"):
                if isinstance(settings.cn_colors, dict):
                    c = settings.cn_colors.get(
                        j, cn_lay.get(j, {"color": (10, 10, 10)})["color"]
                    )
                elif settings.cn_colors == "neuron":
                    c = color
                elif settings.cn_colors:
                    c = settings.cn_colors
                else:
                    c = cn_lay.get(j, {"color": (10, 10, 10)})["color"]

                c = color_to_int(eval_color(c, color_range=255))

                cn_label = f'{cn_lay.get(j, {"name": "connector"})["name"]} of {name}'

                if cn_lay["display"] == "circles" or not isinstance(
                    neuron, core.TreeNeuron
                ):
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore")
                        trace_data.append(
                            k3d.points(
                                positions=this_cn[["x", "y", "z"]].values,
                                name=cn_label,
                                shader="flat",
                                point_size=settings.cn_size
                                if settings.cn_size
                                else cn_lay["size"] * 50,
                                color=c,
                            )
                        )
                elif cn_lay["display"] == "lines":
                    # Find associated treenodes
                    co1 = this_cn[["x", "y", "z"]].values
                    co2 = (
                        neuron.nodes.set_index("node_id")
                        .loc[this_cn.node_id.values, ["x", "y", "z"]]
                        .values
                    )

                    coords = np.array(
                        [
                            co
                            for seg in zip(
                                co1, co1, co2, co2, [[np.nan] * 3] * len(co1)
                            )
                            for co in seg
                        ]
                    )

                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore")
                        trace_data.append(
                            k3d.line(
                                coords,
                                color=c,
                                name=cn_label,
                                width=settings.linewidth,
                                shader="thick",
                            )
                        )
                else:
                    raise ValueError(
                        f'Unknown display type for connectors "{cn_lay["display"]}"'
                    )

    return trace_data


def mesh2k3d(neuron, legendgroup, showlegend, label, color, settings):
    """Convert MeshNeuron to k3d object."""
    # Skip empty neurons
    if neuron.n_vertices == 0:
        return []

    opacity = 1
    if isinstance(color, np.ndarray) and color.ndim == 2:
        if len(color) == len(neuron.vertices):
            color = [color_to_int(c) for c in color]
            color_kwargs = dict(colors=color)
        else:
            raise ValueError("Colors must match number of vertices for K3D meshes.")
    else:
        c = color_to_int(color)
        color_kwargs = dict(color=c)

        if len(color) == 4:
            opacity = color[3]

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        trace_data = [
            k3d.mesh(
                vertices=neuron.vertices.astype("float32"),
                indices=neuron.faces.astype("uint32"),
                name=label,
                flat_shading=False,
                opacity=opacity,
                **color_kwargs,
            )
        ]

    return trace_data


def voxel2k3d(neuron, legendgroup, showlegend, label, color, settings):
    """Convert VoxelNeuron to k3d object."""
    # Skip empty neurons
    if min(neuron.shape) == 0:
        return []

    img = neuron.grid
    if img.dtype not in (np.float32, np.float64):
        img = img.astype(np.float32)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        trace_data = [
            k3d.volume(
                img.T,
                bounds=neuron.bbox.flatten(),
                interpolation=False,
            )
        ]

    return trace_data


def skeleton2k3d(neuron, legendgroup, showlegend, label, color, settings):
    """Convert skeleton (i.e. TreeNeuron) to plotly line plot."""
    if neuron.nodes.empty:
        logger.warning(f"Skipping TreeNeuron w/o nodes: {neuron.label}")
        return []
    elif neuron.nodes.shape[0] == 1:
        logger.warning(f"Skipping single-node skeleton: {neuron.label}")
        return []

    coords = segments_to_coords(neuron)

    # We have to add (None, None, None) to the end of each segment to
    # make that line discontinuous. For reasons I don't quite understand
    # we have to also duplicate the first and the last coordinate in each segment
    # (possibly a bug)
    coords = np.concatenate(
        [co for seg in coords for co in [seg[:1], seg, seg[-1:], [[None] * 3]]], axis=0
    )

    color_kwargs = {}
    if isinstance(color, np.ndarray) and color.ndim == 2:
        # Change colors to rgb integers
        c = [color_to_int(c) for c in color]

        # Next we have to make colors match the segments in `coords`
        c = np.asarray(c)
        ix = dict(zip(neuron.nodes.node_id.values, np.arange(neuron.n_nodes)))
        # Construct sequence node IDs just like we did in `coords`
        # (note that we insert a "-1" for breaks between segments)
        seg_ids = [
            co for seg in neuron.segments for co in [seg[:1], seg, seg[-1:], [-1]]
        ]
        seg_ids = np.concatenate(seg_ids, axis=0)
        # Translate to node indices
        seg_ix = [ix.get(n, 0) for n in seg_ids]

        # Now map this to vertex colors
        seg_colors = [c[i] for i in seg_ix]

        color_kwargs["colors"] = seg_colors
    else:
        color_kwargs["color"] = c = color_to_int(color)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        trace_data = [
            k3d.line(
                coords,
                width=settings.linewidth,
                shader="thick",
                name=label,
                **color_kwargs,
            )
        ]

    # Add soma(s):
    soma = utils.make_iterable(neuron.soma)
    if settings.soma:
        # If soma detection is messed up we might end up producing
        # hundrets of soma which will freeze the session
        if len(soma) >= 10:
            logger.warning(
                f"Neuron {neuron.id} appears to have {len(soma)} "
                "somas. That does not look right - will ignore "
                "them for plotting."
            )
        else:
            for s in soma:
                # Skip `None` somas
                if isinstance(s, type(None)):
                    continue

                # If we have colors for every vertex, we need to find the
                # color that corresponds to this root (or it's parent to be
                # precise)
                if isinstance(c, (list, np.ndarray)):
                    s_ix = np.where(neuron.nodes.node_id == s)[0][0]
                    soma_color = int(c[s_ix])
                else:
                    soma_color = int(c)

                n = neuron.nodes.set_index("node_id").loc[s]
                r = (
                    getattr(n, neuron.soma_radius)
                    if isinstance(neuron.soma_radius, str)
                    else neuron.soma_radius
                )

                sp = tm.primitives.Sphere(radius=r, subdivisions=2)

                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    trace_data.append(
                        k3d.mesh(
                            vertices=sp.vertices
                            + n[["x", "y", "z"]].values.astype("float32"),
                            indices=sp.faces.astype("uint32"),
                            color=soma_color,
                            flat_shading=False,
                            name=f"soma of {label}",
                        )
                    )

    return trace_data


def scatter2k3d(x, **kwargs):
    """Convert DataFrame with x,y,z columns to plotly scatter plot."""
    c = eval_color(
        kwargs.get("color", kwargs.get("c", (100, 100, 100))), color_range=255
    )
    c = color_to_int(c)
    s = kwargs.get("size", kwargs.get("s", 1))
    name = kwargs.get("name", None)

    trace_data = []
    for scatter in x:
        if isinstance(scatter, pd.DataFrame):
            if not all([c in scatter.columns for c in ["x", "y", "z"]]):
                raise ValueError("DataFrame must have x, y and z columns")
            scatter = [["x", "y", "z"]].values

        if not isinstance(scatter, np.ndarray):
            scatter = np.array(scatter)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            trace_data.append(
                k3d.points(
                    positions=scatter, name=name, color=c, point_size=s, shader="dot"
                )
            )
    return trace_data


def dotprops2k3d(x, legendgroup, showlegend, label, color, settings):
    """Convert Dotprops to plotly graph object."""
    tn = x.to_skeleton(scale_vec=settings.dps_scale_vec)

    return skeleton2k3d(tn, legendgroup, showlegend, label, color, settings=settings)


def volume2k3d(x, colormap, settings):
    """Convert Volumes to plotly objects."""
    trace_data = []
    for i, v in enumerate(x):
        # Skip empty data
        if isinstance(v.vertices, np.ndarray):
            if not v.vertices.any():
                continue
        elif not v.vertices:
            continue

        name = getattr(v, "name", None)

        c = colormap[i]
        if len(c) == 4:
            opacity = c[3]
        else:
            opacity = 0.5

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            trace_data = [
                k3d.mesh(
                    vertices=v.vertices.astype("float32"),
                    indices=v.faces.astype("uint32"),
                    name=name,
                    color=color_to_int(c[:3]),
                    flat_shading=False,
                    opacity=opacity,
                )
            ]

    return trace_data



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import csv
import datetime
import io
import json

import pandas as pd

from pathlib import Path
from textwrap import dedent
from typing import List, Union, Iterable, Dict, Optional, Any, TextIO, IO
from urllib3 import HTTPResponse

from .. import config, utils, core
from . import base

__all__ = ["SwcReader", "read_swc", "write_swc"]

# Set up logging
logger = config.get_logger(__name__)

NODE_COLUMNS = ("node_id", "label", "x", "y", "z", "radius", "parent_id")
COMMENT = "#"
DEFAULT_DELIMITER = " "
DEFAULT_PRECISION = 32
DEFAULT_FMT = "{name}.swc"
NA_VALUES = [None, "None"]


class SwcReader(base.BaseReader):
    def __init__(
        self,
        connector_labels: Optional[Dict[str, Union[str, int]]] = None,
        soma_label: Union[str, int] = 1,
        delimiter: str = DEFAULT_DELIMITER,
        precision: int = DEFAULT_PRECISION,
        read_meta: bool = False,
        fmt: str = DEFAULT_FMT,
        errors: str = "raise",
        attrs: Optional[Dict[str, Any]] = None,
    ):
        if not fmt.endswith(".swc"):
            raise ValueError('`fmt` must end with ".swc"')

        super().__init__(
            fmt=fmt, attrs=attrs, file_ext=".swc", errors=errors, name_fallback="SWC"
        )
        self.connector_labels = connector_labels or dict()
        self.soma_label = soma_label
        self.delimiter = delimiter
        self.read_meta = read_meta

        int_, float_ = base.parse_precision(precision)
        self._dtypes = {
            "node_id": int_,
            "parent_id": int_,
            "label": "category",
            "x": float_,
            "y": float_,
            "z": float_,
            "radius": float_,
        }

    @base.handle_errors
    def read_buffer(
        self, f: IO, attrs: Optional[Dict[str, Any]] = None
    ) -> "core.TreeNeuron":
        """Read buffer into a TreeNeuron.

        Parameters
        ----------
        f :         IO
                    Readable buffer (if bytes, interpreted as utf-8).
        attrs :     dict | None
                    Arbitrary attributes to include in the TreeNeuron.

        Returns
        -------
        core.TreeNeuron
        """
        if isinstance(f, HTTPResponse):
            f = io.StringIO(f.data.decode())

        if isinstance(f.read(0), bytes):
            f = io.TextIOWrapper(f, encoding="utf-8")

        header_rows = read_header_rows(f)
        try:
            nodes = pd.read_csv(
                f,
                delimiter=self.delimiter,
                skipinitialspace=True,
                skiprows=len(header_rows),
                comment=COMMENT,
                header=None,
                na_values=NA_VALUES,
            )
            if len(nodes.columns) < len(NODE_COLUMNS):
                raise ValueError("Not enough columns in SWC file.")
            elif len(nodes.columns) > len(NODE_COLUMNS):
                logger.warning(
                    f"Found {len(nodes.columns)} instead of the expected 7 "
                    "columns in SWC file. Assuming additional columns are "
                    "custom properties. You can silence this warning by setting "
                    "`navis.set_loggers('ERROR')`."
                )
                nodes.columns = (
                    list(NODE_COLUMNS) + nodes.columns[len(NODE_COLUMNS) :].tolist()
                )
            else:
                nodes.columns = NODE_COLUMNS
        except pd.errors.EmptyDataError:
            # If file is totally empty, return an empty neuron
            # Note that the TreeNeuron will still complain but it's a better
            # error message
            nodes = pd.DataFrame(columns=NODE_COLUMNS)

        # Check for row with JSON-formatted meta data
        # Expected format '# Meta: {"id": "12345"}'
        if self.read_meta:
            meta_row = [r for r in header_rows if r.lower().startswith("# meta:")]
            if meta_row:
                meta_data = json.loads(meta_row[0][7:].strip())
                attrs = base.merge_dicts(meta_data, attrs)

        return self.read_dataframe(
            nodes, base.merge_dicts({"swc_header": "\n".join(header_rows)}, attrs)
        )

    @base.handle_errors
    def read_dataframe(
        self, nodes: pd.DataFrame, attrs: Optional[Dict[str, Any]] = None
    ) -> "core.TreeNeuron":
        """Convert a SWC-like DataFrame into a TreeNeuron.

        Parameters
        ----------
        nodes :     pandas.DataFrame
        attrs :     dict or None
                    Arbitrary attributes to include in the TreeNeuron.

        Returns
        -------
        core.TreeNeuron
        """
        n = core.TreeNeuron(
            sanitise_nodes(nodes.astype(self._dtypes, errors="ignore", copy=False)),
            connectors=self._extract_connectors(nodes),
        )

        if self.soma_label is not None:
            is_soma_node = n.nodes.label.values == self.soma_label
            if any(is_soma_node):
                n.soma = n.nodes.node_id.values[is_soma_node][0]

        attrs = self._make_attributes({"name": "SWC", "origin": "DataFrame"}, attrs)

        # SWC is special - we do not want to register it
        n.swc_header = attrs.pop("swc_header", "")

        # Try adding properties one-by-one. If one fails, we'll keep track of it
        # in the `.meta` attribute
        meta = {}
        for k, v in attrs.items():
            try:
                n._register_attr(k, v)
            except (AttributeError, ValueError, TypeError):
                meta[k] = v

        if meta:
            n.meta = meta

        return n

    def _extract_connectors(self, nodes: pd.DataFrame) -> Optional[pd.DataFrame]:
        """Infer outgoing/incoming connectors from node labels.

        Parameters
        ----------
        nodes :     pd.DataFrame

        Returns
        -------
        Optional[pd.DataFrame]
                    With columns `["node_id", "x", "y", "z", "connector_id", "type"]`
        """
        if not self.connector_labels:
            return None

        to_concat = [
            pd.DataFrame([], columns=["node_id", "connector_id", "type", "x", "y", "z"])
        ]
        for name, val in self.connector_labels.items():
            cn = nodes[nodes.label == val][["node_id", "x", "y", "z"]].copy()
            cn["connector_id"] = None
            cn["type"] = name
            to_concat.append(cn)

        return pd.concat(to_concat, axis=0)


def sanitise_nodes(nodes: pd.DataFrame, allow_empty=True) -> pd.DataFrame:
    """Check that nodes dataframe is non-empty and is not missing any data.

    Parameters
    ----------
    nodes : pandas.DataFrame

    Returns
    -------
    pandas.DataFrame
    """
    if not allow_empty and nodes.empty:
        raise ValueError("No data found in SWC.")

    is_na = nodes[["node_id", "parent_id", "x", "y", "z"]].isna().any(axis=1)

    if is_na.any():
        # Remove nodes with missing data
        nodes = nodes.loc[~is_na.any(axis=1)]

        # Because we removed nodes, we'll have to run a more complicated root
        # detection
        nodes.loc[~nodes.parent_id.isin(nodes.node_id), "parent_id"] = -1

    return nodes


def read_header_rows(f: TextIO):
    f"""Read {COMMENT}-prefixed lines from the start of a buffer,
    then seek back to the start of the buffer.

    Parameters
    ----------
    f : io.TextIO

    Returns
    -------
    list : List of strings
    """
    out = []
    for line in f:
        if not line.startswith(COMMENT):
            break
        out.append(line)

    f.seek(0)
    return out


def read_swc(
    f: Union[str, pd.DataFrame, Iterable],
    connector_labels: Optional[Dict[str, Union[str, int]]] = {},
    soma_label: Union[str, int] = 1,
    include_subdirs: bool = False,
    delimiter: str = " ",
    parallel: Union[bool, int] = "auto",
    precision: int = 32,
    fmt: str = "{name}.swc",
    read_meta: bool = True,
    limit: Optional[int] = None,
    errors: str = "raise",
    **kwargs,
) -> "core.NeuronObject":
    """Create Neuron/List from SWC file.

    This import is following format specified
    [here](http://www.neuronland.org/NLMorphologyConverter/MorphologyFormats/SWC/Spec.html).

    Parameters
    ----------
    f :                 str | pandas.DataFrame | list thereof
                        Filename, folder, SWC string, URL or DataFrame:
                         - if folder, will import all `.swc` files
                         - if a `.zip`, `.tar` or `.tar.gz` archive will read all
                           SWC files from the file
                         - if a URL (http:// or https://), will download the
                           file and import it
                         - FTP address (ftp://) can point to a folder or a single
                           file
                         - DataFrames are interpreted as a SWC tables
                        See also `limit` parameter to read only a subset of files.
    connector_labels :  dict, optional
                        If provided will extract connectors from SWC.
                        Dictionary must map types to labels:
                        `{'presynapse': 7, 'postsynapse': 8}`
    include_subdirs :   bool, optional
                        If True and `f` is a folder, will also search
                        subdirectories for `.swc` files.
    delimiter :         str
                        Delimiter to use. Passed to `pandas.read_csv`.
    parallel :          "auto" | bool | int
                        Whether to use parallel processes for reading:
                         - "auto" (default): will use parallel processing if
                            more than 200 SWCs are imported.
                          - Integers will be interpreted as the number of
                            processes to use. Defaults to `os.cpu_count() // 2`.
                          - False will use a single process.
                        Ignored for tar archives. Please note that spawning
                        processes incurs an overhead and might not be faster
                        for small numbers of files.
    precision :         int [8, 16, 32, 64] | None
                        Precision for data. Defaults to 32 bit integers/floats.
                        If `None` will let pandas infer data types - this
                        typically leads to higher than necessary precision.
    fmt :               str
                        Formatter to specify how filenames are parsed into
                        neuron attributes. Some illustrative examples:

                          - `{name}.swc` (default) uses the filename
                            (minus the suffix) as the neuron's name property
                          - `{id}.swc` uses the filename as the neuron's ID
                            property
                          - `{name,id}.swc` uses the filename as the neuron's
                            name and ID properties
                          - `{name}.{id}.swc` splits the filename at a "."
                            and uses the first part as name and the second as ID
                          - `{name,id:int}.swc` same as above but converts
                            into integer for the ID
                          - `{name}_{myproperty}.swc` splits the filename at
                            "_" and uses the first part as name and as a
                            generic "myproperty" property
                          - `{name}_{}_{id}.swc` splits the filename at
                            "_" and uses the first part as name and the last as
                            ID. The middle part is ignored.

                        Throws a ValueError if pattern can't be found in
                        filename. Ignored for DataFrames.
    read_meta :         bool
                        If True and SWC header contains a line with JSON-encoded
                        meta data e.g. (`# Meta: {'id': 123}`), these data
                        will be read as neuron properties. `fmt` still takes
                        precedence. Will try to assign meta data directly as
                        neuron attribute (e.g. `neuron.id`). Failing that
                        (can happen for properties intrinsic to `TreeNeurons`),
                        will add a `.meta` dictionary to the neuron.
    limit :             int | str | slice | list, optional
                        When reading from a folder or archive you can use this parameter to
                        restrict the which files read:
                         - if an integer, will read only the first `limit` SWC files
                          (useful to get a sample from a large library of skeletons)
                         - if a string, will interpret it as filename (regex) pattern
                           and only read files that match the pattern; e.g. `limit='.*_R.*'`
                           will only read files that contain `_R` in their filename
                         - if a slice (e.g. `slice(10, 20)`) will read only the files in
                           that range
                         - a list is expected to be a list of filenames to read from
                           the folder/archive
    errors :            "raise" | "log" | "ignore"
                        If "log" or "ignore", errors will not be raised and the
                        mesh will be skipped. Can result in empty output.
    **kwargs
                        Keyword arguments passed to the construction of
                        `navis.TreeNeuron`. You can use this to e.g. set
                        meta data.

    Returns
    -------
    navis.TreeNeuron
                        Contains SWC file header as `.swc_header` attribute.
    navis.NeuronList
                        If import of multiple SWCs will return NeuronList of
                        TreeNeurons.

    See Also
    --------
    [`navis.write_swc`][]
                        Export neurons as SWC files.

    Examples
    --------

    Read a single file:

    >>> s = navis.read_swc('skeleton.swc')                      # doctest: +SKIP

    Read all .swc files in a directory:

    >>> s = navis.read_swc('/some/directory/')                  # doctest: +SKIP

    Read all .swc files in a zip archive:

    >>> s = navis.read_swc('skeletons.zip')                     # doctest: +SKIP

    Sample the first 100 SWC files in a zip archive:

    >>> s = navis.read_swc('skeletons.zip', limit=100)          # doctest: +SKIP

    Read first all SWC files an ftp folder:

    >>> s = navis.read_swc('ftp://server:port/path/to/swc/')    # doctest: +SKIP

    """
    # SwcReader will try its best to read whatever you throw at it - with limited
    # sanity checks. For example: if you misspell a filepath, it will assume
    # that it's a SWC string (because anything that's a string but doesn't
    # point to an existing file or a folder MUST be a SWC) which will lead to
    # strange error messages.
    # The easiest fix is to implement a small sanity check here:
    if isinstance(f, str) and "\n" not in f and not utils.is_url(f):
        # If this looks like a path
        p = Path(f).expanduser()
        if not p.is_dir() and not p.is_file():
            raise FileNotFoundError(
                f'"{f}" looks like a directory or filepath '
                "but does not appear to exist."
            )

    reader = SwcReader(
        connector_labels=connector_labels,
        soma_label=soma_label,
        delimiter=delimiter,
        precision=precision,
        read_meta=read_meta,
        fmt=fmt,
        errors=errors,
        attrs=kwargs,
    )
    res = reader.read_any(f, include_subdirs, parallel, limit=limit)

    failed = []
    for n in core.NeuronList(res):
        if not hasattr(n, "meta"):
            continue
        failed += list(n.meta.keys())

    if failed:
        failed = list(set(failed))
        logger.warning(
            "Some meta data could not be directly attached to the "
            "neuron(s) - probably some clash with intrinsic "
            "properties. You can find these data attached as "
            "`.meta` dictionary."
        )

    return res


def write_swc(
    x: "core.NeuronObject",
    filepath: Union[str, Path],
    header: Optional[str] = None,
    write_meta: Union[bool, List[str], dict] = True,
    labels: Union[str, dict, bool] = True,
    export_connectors: bool = False,
    return_node_map: bool = False,
) -> None:
    """Write TreeNeuron(s) to SWC.

    Follows the format specified
    [here](http://www.neuronland.org/NLMorphologyConverter/MorphologyFormats/SWC/Spec.html).

    Parameters
    ----------
    x :                 TreeNeuron | NeuronList
                        If multiple neurons, will generate a single SWC file
                        for each neuron (see also `filepath`).
    filepath :          str | pathlib.Path | list thereof
                        Destination for the SWC files. See examples for options.
                        If `x` is multiple neurons, `filepath` must either
                        be a folder, a "formattable" filename, a filename ending
                        in `.zip` or a list of filenames (one for each neuron
                        in `x`). Existing files will be overwritten!
    header :            str | None, optional
                        Header for SWC file. If not provided, will use generic
                        header.
    write_meta :        bool | list | dict
                        If not False, will add meta data as JSON-formatted
                        string to the header::

                           True: adds neuron `id`, `name` and `units`
                           list: use to set which properties, e.g. ['id', 'units']
                           dict: use to set meta data, e.g. {'template': 'JRC2018F'}

                        This parameter is ignored if custom header is provided.
    labels :            str | dict | bool, optional
                        Node labels. Can be::

                            str : column name in node table
                            dict: must be of format {node_id: 'label', ...}.
                            bool: if True, will generate automatic labels, if False all nodes have label "0".

    export_connectors : bool, optional
                        If True, will label nodes with pre- ("7") and
                        postsynapse ("8"). Because only one label can be given
                        this might drop synapses (i.e. in case of multiple
                        pre- and/or postsynapses on a single node)! `labels`
                        must be `True` for this to have any effect.
    return_node_map :   bool
                        If True, will return a dictionary mapping the old node
                        ID to the new reindexed node IDs in the file.

    Returns
    -------
    node_map :          dict
                        Only if `return_node_map=True`.

    See Also
    --------
    [`navis.read_swc`][]
                        Import skeleton from SWC files.

    Examples
    --------
    Save a single neuron to a specific file:

    >>> import navis
    >>> n = navis.example_neurons(1, kind='skeleton')
    >>> navis.write_swc(n, tmp_dir / 'my_neuron.swc')

    Save two neurons to specific files:

    >>> import navis
    >>> nl = navis.example_neurons(2, kind='skeleton')
    >>> navis.write_swc(nl, [tmp_dir / 'my_neuron1.swc', tmp_dir / 'my_neuron2.swc'])

    Save multiple neurons to a folder (must exist). Filenames will be
    autogenerated as "{neuron.id}.swc":

    >>> import navis
    >>> nl = navis.example_neurons(5, kind='skeleton')
    >>> navis.write_swc(nl, tmp_dir)

    Save multiple neurons to a folder but modify the pattern for the
    autogenerated filenames:

    >>> import navis
    >>> nl = navis.example_neurons(5, kind='skeleton')
    >>> navis.write_swc(nl, tmp_dir / 'skel-{neuron.name}.swc')

    Save multiple neurons to a zip file:

    >>> import navis
    >>> nl = navis.example_neurons(5, kind='skeleton')
    >>> navis.write_swc(nl, tmp_dir / 'neuronlist.zip')

    Save multiple neurons to a zip file but modify the filenames:

    >>> import navis
    >>> nl = navis.example_neurons(5, kind='skeleton')
    >>> navis.write_swc(nl, tmp_dir / 'skel-{neuron.name}.swc@neuronlist.zip')

    """
    # Make sure inputs are only TreeNeurons
    if isinstance(x, core.NeuronList):
        for n in x:
            if not isinstance(n, core.TreeNeuron):
                msg = f'Can only write TreeNeurons to SWC, not "{type(n)}"'
                if isinstance(n, core.Dotprops):
                    msg += (
                        ". For Dotprops, you can use either `navis.write_nrrd`"
                        " or `navis.write_parquet`."
                    )
                raise TypeError(msg)
    elif not isinstance(x, core.TreeNeuron):
        msg = f'Can only write TreeNeurons to SWC, not "{type(n)}"'
        if isinstance(n, core.Dotprops):
            msg += (
                ". For Dotprops, you can use either `navis.write_nrrd`"
                " or `navis.write_parquet`."
            )
        raise TypeError(msg)

    writer = base.Writer(write_func=_write_swc, ext=".swc")

    return writer.write_any(
        x,
        filepath=filepath,
        header=header,
        write_meta=write_meta,
        labels=labels,
        export_connectors=export_connectors,
        return_node_map=return_node_map,
    )


def _write_swc(
    x: Union["core.TreeNeuron", "core.Dotprops"],
    filepath: Union[str, Path],
    header: Optional[str] = None,
    write_meta: Union[bool, List[str], dict] = True,
    labels: Union[str, dict, bool] = True,
    export_connectors: bool = False,
    return_node_map: bool = False,
) -> None:
    """Write single TreeNeuron to file."""
    # Generate SWC table
    res = make_swc_table(
        x,
        labels=labels,
        export_connectors=export_connectors,
        return_node_map=return_node_map,
    )

    if return_node_map:
        swc, node_map = res[0], res[1]
    else:
        swc = res

    # Generate header if not provided
    if not isinstance(header, str):
        header = dedent(f"""\
        # SWC format file
        # based on specifications at http://www.neuronland.org/NLMorphologyConverter/MorphologyFormats/SWC/Spec.html
        # Created on {datetime.date.today()} using navis (https://github.com/navis-org/navis)
        """)
        if write_meta:
            if isinstance(write_meta, str):
                props = {write_meta: str(getattr(x, write_meta, None))}
            elif isinstance(write_meta, dict):
                props = write_meta
            elif isinstance(write_meta, list):
                props = {k: str(getattr(x, k, None)) for k in write_meta}
            else:
                props = {k: str(getattr(x, k, None)) for k in ["id", "name", "units"]}
            header += f"# Meta: {json.dumps(props)}\n"
        header += dedent("""\
        # PointNo Label X Y Z Radius Parent
        # Labels:
        # 0 = undefined, 1 = soma, 5 = fork point, 6 = end point
        """)
        if export_connectors:
            header += dedent("""\
            # 7 = presynapses, 8 = postsynapses
            """)
    elif not header.endswith("\n"):
        header += "\n"

    with open(filepath, "w") as file:
        # Write header
        file.write(header)

        # Write data
        writer = csv.writer(file, delimiter=" ")
        writer.writerows(swc.astype(str).values)

    if return_node_map:
        return node_map


def make_swc_table(
    x: Union["core.TreeNeuron", "core.Dotprops"],
    labels: Union[str, dict, bool] = None,
    export_connectors: bool = False,
    return_node_map: bool = False,
) -> pd.DataFrame:
    """Generate a node table compliant with the SWC format.

    Follows the format specified
    [here](http://www.neuronland.org/NLMorphologyConverter/MorphologyFormats/SWC/Spec.html).

    Parameters
    ----------
    x :                 TreeNeuron | Dotprops
                        Dotprops will be turned from points + vectors to
                        individual segments.
    labels :            str | dict | bool, optional
                        Node labels. Can be::

                        str : column name in node table
                        dict: must be of format {node_id: 'label', ...}.
                        bool: if True, will generate automatic labels, if False all nodes have label "0".

    export_connectors : bool, optional
                        If True, will label nodes with pre- ("7") and
                        postsynapse ("8"). Because only one label can be given
                        this might drop synapses (i.e. in case of multiple
                        pre- or postsynapses on a single node)! `labels`
                        must be `True` for this to have any effect.
    return_node_map :   bool
                        If True, will return a dictionary mapping the old node
                        ID to the new reindexed node IDs in the file.

    Returns
    -------
    SWC table :         pandas.DataFrame
    node map :          dict
                        Only if `return_node_map=True`.

    """
    if isinstance(x, core.Dotprops):
        x = x.to_skeleton()

    # Work on a copy
    swc = x.nodes.copy()

    # Add labels
    swc["label"] = 0
    if isinstance(labels, dict):
        swc["label"] = swc.index.map(labels)
    elif isinstance(labels, str):
        swc["label"] = swc[labels]
    elif labels:
        # Add end/branch labels
        swc.loc[swc.type == "branch", "label"] = 5
        swc.loc[swc.type == "end", "label"] = 6
        # Add soma label
        if not isinstance(x.soma, type(None)):
            soma = utils.make_iterable(x.soma)
            swc.loc[swc.node_id.isin(soma), "label"] = 1
        if export_connectors:
            # Add synapse label
            pre_ids = x.presynapses.node_id.values
            post_ids = x.postsynapses.node_id.values
            swc.loc[swc.node_id.isin(pre_ids), "label"] = 7
            swc.loc[swc.node_id.isin(post_ids), "label"] = 8

    # Sort such that the parent is always before the child
    swc.sort_values("parent_id", ascending=True, inplace=True)

    # Reset index
    swc.reset_index(drop=True, inplace=True)

    # Generate mapping
    new_ids = dict(zip(swc.node_id.values, swc.index.values + 1))

    swc["node_id"] = swc.node_id.map(new_ids)
    # Lambda prevents potential issue with missing parents
    swc["parent_id"] = swc.parent_id.map(lambda x: new_ids.get(x, -1))

    # Get things in order
    swc = swc[["node_id", "label", "x", "y", "z", "radius", "parent_id"]]

    # Make sure radius has no `None`
    swc["radius"] = swc.radius.fillna(0)

    # Adjust column titles
    swc.columns = ["PointNo", "Label", "X", "Y", "Z", "Radius", "Parent"]

    if return_node_map:
        return swc, new_ids

    return swc



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import numpy as np
import trimesh as tm

from .. import core, utils


def simplify_mesh_open3d(x, F, method='quadric', inplace=False, **kwargs):
    """Simplify mesh using open3d.

    Parameters
    ----------
    x :         MeshNeuron | Volume | Trimesh
                Mesh object to simplify.
    F :         float | int
                For method `quadric` this is the target face count (integer).
                For method `cluster` this is the size of the voxel within which
                vertices are pooled (larger t = coarser mesh).
    method :    "quadric" | "cluster"
                Which method to use for simplification: either Quadric Error
                Metric Decimation (by Garland and Heckbert) or vertex clustering.
                Note that the intepretation of `F` depends on the method.
    inplace :   bool
                If True, will perform simplication on `x`. If False, will
                simplify and return a copy.
    **kwargs
                Keyword arguments are passed through to open3d's
                `simplify_quadric_decimation` and `simplify_vertex_clustering`,
                respectively.

    Returns
    -------
    simp
                Simplified mesh object.

    """
    if not utils.is_mesh(x):
        raise TypeError(f'Expected mesh-like, got "{type(x)}"')

    mesh_o3d = make_o3d_mesh(x)

    if method == 'quadric':
        result = mesh_o3d.simplify_quadric_decimation(int(F), **kwargs)
    elif method == 'cluster':
        result = mesh_o3d.simplify_vertex_clustering(F,  **kwargs)
    else:
        raise ValueError(f'Unknown simplification scheme "{method}"')

    if not inplace:
        x = x.copy()

    x.vertices = np.asarray(result.vertices)
    x.faces = np.asarray(result.triangles)

    return x


def smooth_mesh_open3d(x, iterations=5, L=0.5, inplace=False):
    """Smooth mesh using open3d's Laplacian smoothing.

    Parameters
    ----------
    x :             MeshNeuron | Volume | Trimesh
                    Mesh object to simplify.
    iterations :    int
                    Round of smoothing to apply.
    L :             float [0-1]
                    Diffusion speed constant lambda. Larger = more aggressive
                    smoothing.
    inplace :       bool
                    If True, will perform simplication on `x`. If False, will
                    simplify and return a copy.

    Returns
    -------
    smoothed
                    Smoothed mesh object.

    """
    if not isinstance(x, (core.MeshNeuron, core.Volume, tm.Trimesh)):
        raise TypeError('Expected MeshNeuron, Volume or trimesh.Trimesh, '
                        f'got "{type(x)}"')

    mesh_o3d = make_o3d_mesh(x)

    if L > 1 or L < 0:
        raise ValueError(f'`L` (lambda) must be between 0 and 1, got "{L}"')

    result = mesh_o3d.filter_smooth_laplacian(iterations, L)

    if not inplace:
        x = x.copy()

    x.vertices = np.asarray(result.vertices)
    x.faces = np.asarray(result.triangles)

    return x


def make_o3d_mesh(x):
    """Turn mesh-like object into an open3d mesh."""
    try:
        import open3d
    except ModuleNotFoundError:
        raise ModuleNotFoundError('Please install open3d: pip3 install open3d')
    except BaseException:
        raise

    return open3d.geometry.TriangleMesh(
                vertices=open3d.utility.Vector3dVector(x.vertices),
                triangles=open3d.utility.Vector3iVector(x.faces))



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import numpy as np
import networkx as nx
import pandas as pd
import scipy.spatial
import scipy.sparse

from typing import Union, Optional, List, Iterable

try:
    import igraph
except ModuleNotFoundError:
    igraph = None

from .. import config, core

# Set up logging
logger = config.get_logger(__name__)

__all__ = sorted(
    [
        "network2nx",
        "network2igraph",
        "neuron2igraph",
        "nx2neuron",
        "edges2neuron",
        "neuron2nx",
        "neuron2KDTree",
        "neuron2tangents",
        "simplify_graph",
    ]
)


def neuron2tangents(x: "core.NeuronObject") -> "core.Dotprops":
    """Turn skeleton(s) into points + tangent vectors.

    This will drop zero-length vectors (i.e when node and parent occupy the
    exact same position).

    Parameters
    ----------
    x :         TreeNeuron | NeuronList

    Returns
    -------
    points :    (N, 3) array
                Midpoints for each child->parent node pair.
    vect :      (N, 3) array
                Normalized child-> parent vectors.
    length :    (N, ) array
                Distance between parent and child

    Examples
    --------
    >>> import navis
    >>> n = navis.example_neurons(1)
    >>> t = navis.neuron2tangents(n)

    """
    if isinstance(x, core.NeuronList):
        return [neuron2tangents(n) for n in x]
    elif not isinstance(x, core.TreeNeuron):
        raise TypeError(f'Expected TreeNeuron/List, got "{type(x)}"')

    # Collect nodes
    nodes = x.nodes[x.nodes.parent_id >= 0]

    # Get child->parent vectors
    parent_locs = (
        x.nodes.set_index("node_id").loc[nodes.parent_id, ["x", "y", "z"]].values
    )
    child_locs = nodes[["x", "y", "z"]].values
    vect = child_locs - parent_locs

    # Get mid point
    points = child_locs + (parent_locs - child_locs) / 2

    # Get length
    length = np.sqrt(np.sum(vect**2, axis=1))

    # Drop zero length points
    points = points[length != 0]
    vect = vect[length != 0]
    length = length[length != 0]

    # Normalize vector
    vect = vect / np.linalg.norm(vect, axis=1).reshape(-1, 1)

    return points, vect, length


def network2nx(
    x: Union[pd.DataFrame, Iterable],
    threshold: Optional[float] = None,
    group_by: Union[dict, None] = None,
) -> nx.DiGraph:
    """Generate NetworkX graph from edge list or adjacency.

    Parameters
    ----------
    x :                 pandas.DataFrame
                        Connectivity information:

                         1. List of edges (columns: 'source', 'target', 'weight')
                         2. Adjacency matrix (pd.DataFrame, rows=sources,
                            columns=targets)

    threshold :         float | int, optional
                        Connections weaker than this will be excluded.
    group_by :          None | dict, optional
                        Provide a dictionary `{group_name: [skid1, skid2, ...]}`
                        to collapse sets of nodes into groups.

    Returns
    -------
    networkx.DiGraph
                        NetworkX representation of the network.

    """
    if isinstance(x, pd.DataFrame):
        present = [c in x.columns for c in ["source", "target", "weight"]]
        if all(present):
            edges = x[["source", "target", "weight"]].values
        else:
            # Assume it's an adjacency matrix
            ix_name = x.index.name if x.index.name else "index"
            edges = (
                x.reset_index(inplace=False, drop=False).melt(id_vars=ix_name).values
            )
    elif isinstance(x, (list, np.ndarray)):
        edges = np.array(x)
    else:
        raise TypeError(f'Expected numpy array or pandas DataFrame, got "{type(x)}"')

    if edges.ndim != 2 or edges.shape[1] != 3:
        raise ValueError(
            "Edges must be (N, 3) array containing source, " "target, weight"
        )

    if not isinstance(threshold, (type(None), bool)):
        edges = edges[edges[:, 2] >= threshold]

    # Generate graph and assign custom properties
    g = nx.DiGraph()
    g.add_weighted_edges_from(edges)

    # Group nodes
    if group_by:
        for n, skids in group_by.items():
            # First collapse all nodes into the first of each group
            for s in skids[1:]:
                g = nx.contracted_nodes(g, str(skids[0]), str(s))
            # Now relabel the first node
            g = nx.relabel_nodes(g, {str(skids[0]): str(n)})
            g.nodes[str(n)]["neuron_name"] = str(n)

    return g


def network2igraph(
    x: Union[pd.DataFrame, Iterable], threshold: Optional[float] = None
) -> "igraph.Graph":
    """Generate iGraph graph from edge list or adjacency.

    Requires iGraph to be installed.

    Parameters
    ----------
    x :                 pandas.DataFrame | np.array
                        Connectivity information:

                         1. List of edges (columns: 'source', 'target', 'weight')
                         2. Adjacency matrix (pd.DataFrame, rows=sources,
                            columns=targets)

    threshold :         float | int, optional
                        Connections weaker than this will be excluded.

    Returns
    -------
    igraph.Graph(directed=True)
                        iGraph representation of the network.

    """
    if igraph is None:
        raise ModuleNotFoundError("igraph must be installed to use this function.")

    if isinstance(x, pd.DataFrame):
        present = [c in x.columns for c in ["source", "target", "weight"]]
        if all(present):
            edges = x[["source", "target", "weight"]].values
        else:
            edges = (
                x.reset_index(inplace=False, drop=False)
                .melt(id_vars="index", inplace=False)
                .values
            )
    elif isinstance(x, (list, np.ndarray)):
        edges = np.array(x)
    else:
        raise TypeError(f'Expected numpy array or pandas DataFrame, got "{type(x)}"')

    if edges.ndim != 2 or edges.shape[1] != 3:
        raise ValueError(
            "Edges must be (N, 3) array containing source, " "target, weight"
        )

    if not isinstance(threshold, (type(None), bool)):
        edges = edges[edges[:, 2] >= threshold]

    names = list(set(np.array(edges)[:, 0]) | set(np.array(edges)[:, 1]))

    edges_by_index = [[names.index(e[0]), names.index(e[1])] for e in edges]

    # Generate igraph and assign custom properties
    g = igraph.Graph(directed=True)
    g.add_vertices(len(names))
    g.add_edges(edges_by_index)

    g.vs["node_id"] = names
    # g.vs['neuron_name'] = g.vs['label'] = neuron_names
    g.es["weight"] = edges[:, 2]

    return g


def neuron2nx(x: "core.NeuronObject", simplify=False, epsilon=None) -> nx.DiGraph:
    """Turn Tree-, Mesh- or VoxelNeuron into an NetworkX graph.

    Parameters
    ----------
    x :         TreeNeuron | MeshNeuron | VoxelNeuron | NeuronList
                Uses simple 6-connectedness for voxels.
    simplify :  bool
                For TreeNeurons only: simplify the graph by keeping only roots,
                leaves and branching points. Preserves the original
                branch lengths (i.e. weights).
    epsilon :   float, optional
                For Dotprops only: maximum distance between two points to
                connect them. If `None`, will use 5x the average distance
                between points (i.e. `5 * x.sampling_resolution`).

    Returns
    -------
    graph:      networkx.Graph | networkx.DiGraph
                NetworkX representation of the neuron. Returns list of graphs
                if x is multiple neurons. Graph is directed for TreeNeurons
                and undirected for Mesh- and VoxelNeurons. Graph is weighted
                for Tree- and MeshNeurons.

    """
    if isinstance(x, core.NeuronList):
        return [neuron2nx(x.loc[i]) for i in range(x.shape[0])]

    if isinstance(x, core.TreeNeuron):
        # Collect nodes
        nodes = x.nodes.set_index("node_id", inplace=False)
        # Collect edges
        edges = x.nodes[x.nodes.parent_id >= 0][["node_id", "parent_id"]].values
        # Collect weight
        weights = np.sqrt(
            np.sum(
                (
                    nodes.loc[edges[:, 0], ["x", "y", "z"]].values.astype(float)
                    - nodes.loc[edges[:, 1], ["x", "y", "z"]].values.astype(float)
                )
                ** 2,
                axis=1,
            )
        )
        # It's fastest to generate a list of (source, target, weight) tuples to pass to networkX
        elist = [(e[0], e[1], l) for e, l in zip(edges, weights)]
        # Create empty directed Graph
        G = nx.DiGraph()
        # Add nodes (in case we have disconnected nodes)
        G.add_nodes_from(x.nodes.node_id.values)
        # Add edges
        G.add_weighted_edges_from(elist)

        if simplify:
            simplify_graph(G, inplace=True)
    elif isinstance(x, core.MeshNeuron):
        G = nx.Graph()
        G.add_nodes_from(np.arange(x.n_vertices))
        edges = [
            (e[0], e[1], l)
            for e, l in zip(x.trimesh.edges_unique, x.trimesh.edges_unique_length)
        ]
        G.add_weighted_edges_from(edges)
    elif isinstance(x, core.Dotprops):
        if epsilon is None:
            epsilon = 5 * x.sampling_resolution

        # Generate KDTree
        tree = neuron2KDTree(x)

        # Generate graph and assign custom properties
        G = nx.Graph()
        G.add_nodes_from(np.arange(x.n_points))
        G.add_edges_from(tree.query_pairs(epsilon))
    elif isinstance(x, core.VoxelNeuron):
        # First we need to determine the 6-connecivity between voxels
        edges = []
        # Go over each axis
        for i in range(3):
            # Generate an offset of 1 voxel along given axis
            offset = np.zeros(3, dtype=int)
            offset[i] = 1
            # Combine real and offset voxels
            vox_off = x.voxels + offset
            # Find out which voxels overlap (i.e. count == 2 after offset)
            unique, cnt = np.unique(
                np.append(x.voxels, vox_off, axis=0), axis=0, return_counts=True
            )

            connected = unique[cnt > 1]
            for vox in connected:
                edges.append([tuple(vox), tuple(vox - offset)])
        G = nx.Graph()
        G.add_nodes_from([tuple(v) for v in x.voxels])
        G.add_edges_from(edges)
    else:
        raise ValueError(
            f'Unable to convert data of type "{type(x)}" to networkx graph.'
        )

    return G


def simplify_graph(G, inplace=False):
    """Simplify skeleton graph (networkX or igraph).

    This function will simplify the graph by keeping only roots, leafs and
    branch points. Preserves branch lengths (i.e. weights)!

    Parameters
    ----------
    G :         networkx.DiGraph | igraph.Graph
                The skeleton graph to simplify.
    inplace :   bool
                If True, will modify the graph in place.

    Returns
    -------
    G :         networkx.DiGraph | networkx.DiGraph
                Simplified graph.

    Examples
    --------
    >>> import navis
    >>> n = navis.example_neurons(1, kind='skeleton')
    >>> # Simplify skeleton's NetworkX graph representation
    >>> G_simp_nx = navis.graph.simplify_graph(n.graph)
    >>> # Check that we have the expected number of nodes
    >>> assert len(G_simp_nx.nodes) == (n.n_branches + n.n_root + n.n_leafs)
    >>> # Simplify skeleton's iGraph graph representation
    >>> G_simp_ig = navis.graph.simplify_graph(n.igraph)
    >>> # Check that we have the expected number of nodes
    >>> assert len(G_simp_ig.vs) == (n.n_branches + n.n_root + n.n_leafs)

    """
    if not inplace:
        G = G.copy()

    if isinstance(G, nx.Graph):
        # Find all leaf and branch points
        leafs = {n for n in G.nodes if G.in_degree(n) == 0 and G.out_degree(n) != 0}
        branches = {n for n in G.nodes if G.in_degree(n) > 1 and G.out_degree(n) != 0}
        roots = {n for n in G.nodes if G.out_degree(n) == 0}

        stop_nodes = roots | leafs | branches

        # Walk from each leaf/branch point to the next leaf, branch or root
        to_remove = []
        for start_node in leafs | branches:
            dist = 0
            node = start_node
            while True:
                parent = next(G.successors(node))
                dist += G.edges[node, parent]["weight"]

                if parent in stop_nodes:
                    G.add_weighted_edges_from([(start_node, parent, dist)])
                    break

                to_remove.append(parent)
                node = parent

        G.remove_nodes_from(to_remove)
    else:
        # Find all leaf and branch points
        leafs = G.vs.select(_indegree=0, _outdegree_ne=0)
        branches = G.vs.select(_indegree_gt=1, _outdegree_ne=0)
        roots = G.vs.select(_outdegree=0)

        stop_nodes = np.concatenate((roots.indices, leafs.indices, branches.indices))

        # Walk from each leaf/branch point to the next leaf, branch or root
        to_remove = []
        for start_node in np.concatenate((leafs.indices, branches.indices)):
            dist = 0
            node = start_node
            while True:
                parent = G.successors(node)[0]
                dist += G.es[G.get_eid(node, parent)]["weight"]

                if parent in stop_nodes:
                    G.add_edge(start_node, parent, weight=dist)
                    break

                to_remove.append(parent)
                node = parent

        G.delete_vertices(to_remove)

    return G


def _voxels2edges(x, connectivity=18):
    """Turn VoxelNeuron into an edges.

    This is function requires scikit-learn to be available.

    Parameters
    ----------
    x :             VoxelNeuron
    connectivity :  6 | 18 | 26
                    Connectedness:
                     - 6 = faces
                     - 18 = faces + edges
                     - 26 = faces + edges + vertices

    Returns
    -------
    edges :         (N, 2) numpy array

    """
    # The distances and metric we will use depend on the connectedness
    METRICS = {6: "manhattan", 18: "euclidean", 26: "chebyshev"}
    DISTANCES = {6: 1, 18: 1.5, 26: 1}

    try:
        from sklearn.neighbors import KDTree
    except ModuleNotFoundError:
        raise ModuleNotFoundError("This function requires scikit-learn to be installed.")

    assert connectivity in (
        6,
        18,
        26,
    ), f'`connectivity` must be 6, 18 or 26, not "{connectivity}"'
    assert isinstance(x, core.VoxelNeuron)

    voxels = x.voxels
    # Create tree with given distance metric
    tree = KDTree(voxels, leaf_size=40, metric=METRICS[connectivity])

    # Query ball pairs
    indices = tree.query_radius(voxels, r=DISTANCES[connectivity])

    # Collected edges
    edges = []
    for i, hits in enumerate(indices):
        # Add edges
        edges += [(i, ix) for ix in hits]
    edges = np.array(edges)

    # Drop self-hits
    edges = edges[edges[:, 0] != edges[:, 1]]

    # Keep only A->B edges and drop B->A edges
    edges = np.unique(np.sort(edges, axis=1), axis=0)

    return edges


def neuron2igraph(
    x: "core.NeuronObject",
    simplify: bool = False,
    connectivity: int = 18,
    raise_not_installed: bool = True,
) -> "igraph.Graph":
    """Turn Tree-, Mesh- or VoxelNeuron(s) into an iGraph graph.

    Requires iGraph to be installed.

    Parameters
    ----------
    x :                     TreeNeuron | MeshNeuron | VoxelNeuron | NeuronList
                            Neuron(s) to convert.
    simplify :              bool
                            For TreeNeurons only: simplify the graph by keeping only roots,
                            leaves and branching points. Preserves the original branch
                            lengths (i.e. weights).
    connectivity :          6 | 18 | 26
                            For VoxelNeurons only. Defines the connectedness:
                             - 6 = faces
                             - 18 = faces + edges
                             - 26 = faces + edges + vertices
    raise_not_installed :   bool
                            If False and igraph is not installed will silently
                            return `None`.

    Returns
    -------
    igraph.Graph
                Representation of the neuron. Returns list of graphs
                if x is multiple neurons. Directed for TreeNeurons, undirected
                for MeshNeurons.
    None
                If igraph not installed.

    """
    # If iGraph is not installed return nothing
    if igraph is None:
        if not raise_not_installed:
            return None
        else:
            raise ModuleNotFoundError(
                "iGraph appears to not be installed (properly). "
                'Make sure "import igraph" works.'
            )

    if isinstance(x, core.NeuronList):
        return [
            neuron2igraph(x.loc[i], connectivity=connectivity)
            for i in range(x.shape[0])
        ]

    if isinstance(x, core.TreeNeuron):
        # Make sure we have correctly numbered indices
        nodes = x.nodes.reset_index(inplace=False, drop=True)

        # Generate list of vertices -> this order is retained
        vlist = nodes.node_id.values

        # Get list of edges as indices (needs to exclude root node)
        tn_index_with_parent = nodes.index.values[nodes.parent_id >= 0]
        parent_ids = nodes.parent_id.values[nodes.parent_id >= 0]
        nodes["temp_index"] = nodes.index  # add temporary index column
        try:
            parent_index = (
                nodes.set_index("node_id", inplace=False)
                .loc[parent_ids, "temp_index"]
                .values
            )
        except KeyError:
            miss = nodes[~nodes.parent_id.isin(nodes.node_id)].node_id.unique()
            raise KeyError(
                f"{len(miss)} nodes (e.g. {miss[0]}) in TreeNeuron "
                f"{x.id} connect to non-existent parent nodes."
            )
        except BaseException:
            raise

        # Generate list of edges based on index of vertices
        elist = np.vstack((tn_index_with_parent, parent_index)).T

        # iGraph < 0.8.0 does not like arrays as edge list
        if getattr(igraph, "__version_info__", (0, 0, 0))[1] < 8:
            elist = elist.tolist()

        # Generate graph and assign custom properties
        G = igraph.Graph(elist, n=len(vlist), directed=True)

        G.vs["node_id"] = G.vs["name"] = nodes.node_id.values
        G.vs["parent_id"] = nodes.parent_id.values

        # Generate weights by calculating edge lengths = distance between nodes
        tn_coords = nodes[["x", "y", "z"]].values[tn_index_with_parent, :]
        parent_coords = nodes[["x", "y", "z"]].values[parent_index.astype(int), :]

        w = np.sqrt(np.sum((tn_coords - parent_coords) ** 2, axis=1))
        G.es["weight"] = w

        if simplify:
            simplify_graph(G, inplace=True)
    elif isinstance(x, core.MeshNeuron):
        elist = x.trimesh.edges_unique
        G = igraph.Graph(elist, n=x.n_vertices, directed=False)
        G.es["weight"] = x.trimesh.edges_unique_length
    elif isinstance(x, core.VoxelNeuron):
        edges = _voxels2edges(x, connectivity=connectivity)
        G = igraph.Graph(edges, n=len(x.voxels), directed=False)
    else:
        raise ValueError(f'Unable to convert data of type "{type(x)}" to igraph.')

    return G


def nx2neuron(
    G: nx.Graph,
    root: Optional[Union[int, str]] = None,
    break_cycles: bool = False,
    **kwargs,
) -> pd.DataFrame:
    """Create TreeNeuron from NetworkX Graph.

    This function will try to generate a neuron-like tree structure from
    the Graph. Therefore the graph must not contain loops!

    All node attributes (e.g. `x`, `y`, `z`, `radius`) will be added to
    the neuron's `.nodes` table.

    Parameters
    ----------
    G :             networkx.Graph
                    Graph to convert to neuron.
    root :          str | int | list, optional
                    Node in graph to use as root for neuron. If not provided,
                    will use first node in `g.nodes`. Ignored if graph
                    consists of several disconnected components.
    break_cycles :  bool
                    The input graph must not contain cycles. We can break them
                    up at risk of disconnecting parts of the graph.
    **kwargs
                    Keyword arguments are passed to the construction of
                    [`navis.TreeNeuron`][].

    Returns
    -------
    TreeNeuron

    Examples
    --------
    >>> import navis
    >>> import networkx as nx
    >>> G = nx.balanced_tree(2, 3)
    >>> tn = navis.nx2neuron(G)
    >>> tn
    type            navis.TreeNeuron
    name                        None
    n_nodes                       15
    n_connectors                None
    n_branches                     6
    n_leafs                        8
    cable_length                 0.0
    soma                        None
    units            1 dimensionless
    dtype: object

    """
    # First some sanity checks
    if not isinstance(G, nx.Graph):
        raise TypeError(f'`G` must be NetworkX Graph, got "{type(G)}"')

    # We need an undirected Graph
    if isinstance(G, nx.DiGraph):
        G = G.to_undirected(as_view=True)

    if not nx.is_forest(G):
        if not break_cycles:
            raise TypeError(
                "Graph must be tree-like. You can try setting "
                "the `cut_cycles` parameter to True."
            )
        else:
            if break_cycles:
                while True:
                    try:
                        # Find cycle
                        cycle = nx.find_cycle(G)
                    except nx.exception.NetworkXNoCycle:
                        break
                    except BaseException:
                        raise

                    # Sort by degree
                    cycle = sorted(cycle, key=lambda x: G.degree[x[0]])

                    # Remove the edge with the lowest degree
                    G.remove_edge(cycle[0][0], cycle[0][1])

    # Ignore root if this is a forest
    if not nx.is_tree(G):
        root = None

    # This effectively makes sure that all edges point in the same direction
    lop = {}
    for c in nx.connected_components(G):
        sg = nx.subgraph(G, c)
        # Pick a random root if not explicitly provided
        if not root:
            r = list(sg.nodes)[0]
        elif root not in sg.nodes:
            raise ValueError(f'Node "{root}" not in graph.')
        else:
            r = root

        # Generate parent->child dictionary
        this_lop = nx.predecessor(sg, r)

        # Make sure no node has more than one parent
        if any((len(v) > 1 for v in this_lop.values())):
            raise ValueError(
                "Nodes with multiple parents found. Make sure graph is tree-like."
            )

        # Note that we assign -1 as root's parent
        lop.update({k: v[0] if v else -1 for k, v in this_lop.items()})

    # Generate node table
    tn_table = pd.DataFrame(index=list(G.nodes))
    tn_table.index = tn_table.index.set_names("node_id", inplace=False)

    # Add parents - use -1 for root's parent
    tn_table["parent_id"] = tn_table.index.map(lop)

    try:
        tn_table.index = tn_table.index.astype(int)
        tn_table["parent_id"] = tn_table.parent_id.astype(int)
    except (ValueError, TypeError):
        raise ValueError("Node IDs must be convertible to integers.")
    except BaseException:
        raise

    # Add additional generic attribute -> will skip node_id and parent_id
    # if they exist
    all_attr = set([k for n in G.nodes for k in G.nodes[n].keys()])

    # Remove some that we don't need
    all_attr -= set(["parent_id", "node_id"])
    # Add some that we want as columns even if they don't exist
    all_attr |= set(["x", "y", "z", "radius"])

    # For some we want to have set default values
    defaults = {"x": 0, "y": 0, "z": 0, "radius": -1}

    # Now map the attributes onto node table
    for at in all_attr:
        vals = nx.get_node_attributes(G, at)
        tn_table[at] = tn_table.index.map(vals).fillna(defaults.get(at, None))

    return core.TreeNeuron(tn_table.reset_index(drop=False, inplace=False), **kwargs)


def edges2neuron(edges, vertices=None, validate=True, **kwargs):
    """Create TreeNeuron from edges and (optional) vertex coordinates.

    Parameters
    ----------
    edges :         (N, 2) array
                    Edges between vertices.
    vertices :      (N, 3) array, optional
                    Vertex positions. If not provided, will position
                    all vertices at (0, 0, 0).
    validate :      bool
                    If True (default) will fix issues with cycles
                    and edges orientation. Only skip this if
                    you are absolutely sure your data are good.
    **kwargs
                    Additional keyword arguments are passed to
                    initialization of the TreeNeuron.

    Returns
    -------
    TreeNeuron

    Examples
    --------

    >>> import navis
    >>> import numpy as np
    >>> verts = np.random.rand(5, 3)
    >>> edges = np.array([(0, 1), (1, 2), (2, 3), (2, 4)])
    >>> sk = navis.edges2neuron(edges, vertices=verts)

    """
    # Make sure we're dealing with arrays
    edges = np.asarray(edges)

    if vertices is not None:
        vertices = np.asarray(vertices)
    else:
        vertices = np.zeros((edges.max() + 1, 3))

    if vertices.ndim != 2 or vertices.shape[1] != 3:
        raise ValueError(
            f"Expected `vertices` to be of shape (N, 3), got {vertices.shape}"
        )
    if edges.ndim != 2 or edges.shape[1] != 2:
        raise ValueError(
            f"Expected `edges` to be of shape (N, 2), got {edges.shape}"
        )

    if edges.max() > (len(vertices)-1):
        raise IndexError("vertex index out of range")

    G = nx.Graph()
    G.add_nodes_from(np.arange(len(vertices)))
    G.add_edges_from(edges)

    # Note: at this point we could just pass the graph to nx2neuron
    # But because we know it came from from vertices and edges, we
    # can skip certain checks and make the process a bit faster

    if validate:
        if not nx.is_forest(G):
            while True:
                try:
                    # Find cycle
                    cycle = nx.find_cycle(G)
                except nx.exception.NetworkXNoCycle:
                    break
                except BaseException:
                    raise

                # Sort by degree
                cycle = sorted(cycle, key=lambda x: G.degree[x[0]])

                # Remove the edge with the lowest degree
                G.remove_edge(cycle[0][0], cycle[0][1])

        parents = {}
        for cc in nx.connected_components(G):
            # If this is a disconnected node
            if len(cc) == 1:
                parents[cc.pop()] = -1
                continue

            sg = nx.subgraph(G, cc)
            # Pick a random root
            r = cc.pop()
            # Generate parent->child dictionary
            this = nx.predecessor(sg, r)

            # Update overall parent dictionary
            # (note that we assign -1 as root's parent)
            parents.update({k: v[0] if v else -1 for k, v in this.items()})

    nodes = pd.DataFrame(vertices, columns=['x', 'y', 'z'])
    nodes.insert(0, 'node_id', nodes.index)
    nodes.insert(1, 'parent_id', nodes.index.map(parents))

    return core.TreeNeuron(nodes, **kwargs)



def _find_all_paths(
    g: nx.DiGraph, start, end, mode: str = "OUT", maxlen: Optional[int] = None
) -> list:
    """Find all paths between two vertices in an iGraph object.

    For some reason this function exists in R iGraph but not Python iGraph. This
    is rather slow and should not be used for large graphs.

    """

    def find_all_paths_aux(
        adjlist: List[set],
        start: int,
        end: int,
        path: list,
        maxlen: Optional[int] = None,
    ) -> list:
        path = path + [start]
        if start == end:
            return [path]
        paths: list = []
        if maxlen is None or len(path) <= maxlen:
            for node in adjlist[start] - set(path):
                paths.extend(find_all_paths_aux(adjlist, node, end, path, maxlen))
        return paths

    adjlist = [set(g.neighbors(node, mode=mode)) for node in range(g.vcount())]
    all_paths: list = []
    start = start if isinstance(start, list) else [start]
    end = end if isinstance(end, list) else [end]
    for s in start:
        for e in end:
            all_paths.extend(find_all_paths_aux(adjlist, s, e, [], maxlen))
    return all_paths


def neuron2KDTree(
    x: "core.NeuronObject", tree_type: str = "c", data: str = "auto", **kwargs
) -> Union[scipy.spatial.cKDTree, scipy.spatial.KDTree]:
    """Turn neuron into scipy KDTree.

    Parameters
    ----------
    x :         TreeNeuron | MeshNeuron | VoxelNeuron | Dotprops
                A single neuron to turn into a KDTree.
    tree_type : 'c' | 'normal'
                Type of KDTree:
                  1. `'c'` = `scipy.spatial.cKDTree` (faster)
                  2. `'normal'` = `scipy.spatial.KDTree` (more functions)
    data :      'auto' | str
                Data used to generate tree. "auto" will pick the core data
                depending on neuron type: `nodes`, `vertices`, `voxels` and
                `points` for TreeNeuron, MeshNeuron, VoxelNeuron and Dotprops,
                respectively. Other values (e.g. "connectors" or "nodes") must
                map to a neuron property that is either (N, 3) array or
                DataFrame with x/y/z columns.
    **kwargs
                Keyword arguments passed at KDTree initialization.


    Returns
    -------
    `scipy.spatial.cKDTree` or `scipy.spatial.KDTree`

    """
    if tree_type not in ["c", "normal"]:
        raise ValueError('"tree_type" needs to be either "c" or "normal"')

    if isinstance(x, core.NeuronList):
        if len(x) == 1:
            x = x[0]
        else:
            raise ValueError("Need a single TreeNeuron")
    elif not isinstance(x, core.BaseNeuron):
        raise TypeError(f'Need Neuron, got "{type(x)}"')

    if data == "auto":
        if isinstance(x, core.TreeNeuron):
            data = "nodes"
        if isinstance(x, core.MeshNeuron):
            data = "vertices"
        if isinstance(x, core.VoxelNeuron):
            data = "voxels"
        if isinstance(x, core.Dotprops):
            data = "points"

    if not hasattr(x, data):
        raise ValueError(f"Neuron does not have a {data} property")

    data = getattr(x, data)

    if isinstance(data, pd.DataFrame):
        if not all(np.isin(["x", "y", "z"], data.columns)):
            raise ValueError(
                f'"{data}" DataFrame must contain "x", "y" and ' '"z" columns.'
            )
        data = data[["x", "y", "z"]].values

    if not isinstance(data, np.ndarray) or data.ndim != 2 or data.shape[1] != 3:
        raise ValueError(
            f'"{data}" must be DataFrame or (N, 3) array, got {type(data)}'
        )

    if tree_type == "c":
        return scipy.spatial.cKDTree(data=data, **kwargs)
    else:
        return scipy.spatial.KDTree(data=data, **kwargs)



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

from .mmetrics import (strahler_index, bending_flow,
                       flow_centrality, synapse_flow_centrality, sholl_analysis,
                       segregation_index, arbor_segregation_index, tortuosity,
                       betweeness_centrality, segment_analysis, cable_length)
from .manipulation import (prune_by_strahler, stitch_skeletons,
                           split_axon_dendrite, average_skeletons,
                           despike_skeleton, guess_radius, smooth_skeleton,
                           heal_skeleton, break_fragments,
                           prune_twigs, prune_at_depth, cell_body_fiber,
                           drop_fluff, combine_neurons)
from .analyze import find_soma
from .subset import subset_neuron
from .persistence import (persistence_points, persistence_vectors,
                          persistence_distances)
from .fq import form_factor
from .ivscc import ivscc_features
from .images import smooth_voxels, thin_voxels


__all__ = ['strahler_index', 'bending_flow', 'flow_centrality', 'synapse_flow_centrality',
           'segregation_index', 'arbor_segregation_index', 'tortuosity',
           'prune_by_strahler', 'stitch_skeletons', 'split_axon_dendrite',
           'average_skeletons', 'despike_skeleton', 'guess_radius', 'smooth_skeleton',
           'heal_skeleton', 'break_fragments', 'prune_twigs',
           'find_soma', 'prune_at_depth', 'cell_body_fiber', 'drop_fluff',
           'subset_neuron', 'smooth_voxels', 'sholl_analysis',
           'persistence_points', 'betweeness_centrality',
           'persistence_vectors', 'persistence_distances', 'combine_neurons',
           'segment_analysis', 'form_factor', 'ivscc_features', "thin_voxels"]



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import pathlib
import json

from .affine import AffineTransform
from .base import TransformSequence
from .cmtk import CMTKtransform
from .h5reg import H5transform
from .thinplate import TPStransform


def parse_json(filepath: str, **kwargs):
    """Parse json-encoded transform (experimental).

    Parameters
    ----------
    filepath :      str
                    Filepath to json file to generate transform from.
    **kwargs
                    Keyword arguments passed to the respective transform class.

    Returns
    -------
    transform

    """
    fp = pathlib.Path(filepath)

    with open(fp, 'r') as f:
        data = json.load(f)

    if not isinstance(data, list):
        data = []

    transforms = []
    for reg in data:
        if not isinstance(reg, type(dict)):
            raise TypeError(f'{filepath} expected data as dict or list of '
                            f'dicts, got "{type(reg)}"')

        if reg.get('type', None) == 'tpsreg':
            transforms.append(TPStransform(reg['refmat'].values,
                                           reg['tarmat'].values, **kwargs))
        elif reg.get('type', None) == 'affine':
            transforms.append(AffineTransform(reg['affine_matrix'], **kwargs))
        else:
            raise TypeError(f'{reg} has unknown "type"')

    return TransformSequence(*transforms)


def transform_factory(filepath: str, **kwargs):
    """Generate appropriate transform from file.

    Parameters
    ----------
    filepath :      str
                    Filepath to generate transform from.
    **kwargs
                    Keyword arguments passed to the respective transform class.

    Returns
    -------
    transform

    """
    fp = pathlib.Path(filepath)

    # Check if file/path exists
    if not fp.is_dir() or not fp.is_file():
        raise ValueError(f'{fp} does not appear to exist')

    if fp.endswith('.list'):
        return CMTKtransform(fp, **kwargs)

    if fp.endswith('.h5'):
        return H5transform(fp, **kwargs)

    # Custom transforms
    if fp.endswith('.json'):
        return parse_json(fp, **kwargs)

    raise TypeError(f'Unknown transform format for {filepath}')


# This must comme after defining/importing the functions
factory_methods = {'.list': CMTKtransform,
                   '.h5': H5transform,
                   '.json': parse_json}
